#ifndef __SPROX_DESFIRE_H__
#define __SPROX_DESFIRE_H__

#include "hardware/springprox/springprox.h"
#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>

#ifndef SPROX_DESFIRE_LIB
#define SPROX_DESFIRE_LIB __declspec( dllimport )
#endif

#ifndef UNDER_CE
    /* Under Desktop Windows we use the cdecl calling convention */
#define SPROX_DESFIRE_API __cdecl
#else
    /* Under Windows CE we use the stdcall calling convention */
#define SPROX_DESFIRE_API __stdcall
#endif

  /* Shell we provide UNICODE functions ? */
#ifdef UNICODE
#define DesFire_GetErrorMessageW DesFire_GetErrorMessage
#else
#define DesFire_GetErrorMessageA DesFire_GetErrorMessage
#endif

#endif

#define APPLICATION_ID_SIZE          3
#define MAX_INFO_FRAME_SIZE          60

#ifndef WIN32
#define SPROX_DESFIRE_LIB
#define SPROX_DESFIRE_API
#endif

/*
 * DESFire status (codes returned by the PICC)
 * -------------------------------------------
 */

#define DF_OPERATION_OK                     0x00
#define DF_NO_CHANGES                       0x0C
#define DF_OUT_OF_EEPROM_ERROR              0x0E
#define DF_ILLEGAL_COMMAND_CODE             0x1C
#define DF_INTEGRITY_ERROR                  0x1E
#define DF_NO_SUCH_KEY                      0x40
#define DF_LENGTH_ERROR                     0x7E
#define DF_PERMISSION_DENIED                0x9D
#define DF_PARAMETER_ERROR                  0x9E
#define DF_APPLICATION_NOT_FOUND            0xA0
#define DF_APPL_INTEGRITY_ERROR             0xA1
#define DF_AUTHENTICATION_CORRECT           0xAC
#define DF_AUTHENTICATION_ERROR             0xAE
#define DF_ADDITIONAL_FRAME                 0xAF
#define DF_BOUNDARY_ERROR                   0xBE
#define DF_CARD_INTEGRITY_ERROR             0xC1
#define DF_COMMAND_ABORTED                  0xCA
#define DF_CARD_DISABLED_ERROR              0xCD
#define DF_COUNT_ERROR                      0xCE
#define DF_DUPLICATE_ERROR                  0xDE
#define DF_FILE_NOT_FOUND                   0xF0
#define DF_FILE_INTEGRITY_ERROR             0xF1
#define DF_EEPROM_ERROR                     0xEE

/*
 * DESFire API status/error codes.
 * -------------------------------
 */
#define DFCARD_ERROR                        -1000
#define DFCARD_LIB_CALL_ERROR               DFCARD_ERROR+1
#define DFCARD_OUT_OF_MEMORY                DFCARD_ERROR+2
#define DFCARD_OVERFLOW                     DFCARD_ERROR+3
#define DFCARD_WRONG_KEY                    DFCARD_ERROR+4
#define DFCARD_WRONG_MAC                    DFCARD_ERROR+5
#define DFCARD_WRONG_CRC                    DFCARD_ERROR+6
#define DFCARD_WRONG_LENGTH                 DFCARD_ERROR+7
#define DFCARD_WRONG_PADDING                DFCARD_ERROR+8
#define DFCARD_WRONG_FILE_TYPE              DFCARD_ERROR+9
#define DFCARD_WRONG_RECORD_SIZE            DFCARD_ERROR+10

/*
 * DESFire file types
 * ------------------
 */
#define	DF_STANDARD_DATA_FILE			          0x00
#define	DF_BACKUP_DATA_FILE				          0x01
#define	DF_VALUE_FILE                       0x02
#define	DF_LINEAR_RECORD_FILE			          0x03
#define	DF_CYCLIC_RECORD_FILE			          0x04

/*
 * DESFire communication mode flag combinations.
 * ---------------------------------------------
 */
#define	DF_COMM_MODE_PLAIN			            0x00
#define	DF_COMM_MODE_MACED			            0x01
#define	DF_COMM_MODE_PLAIN2			            0x02
#define	DF_COMM_MODE_ENCIPHERED		          0x03

/*
 * DESFire access rights
 * ---------------------
 */

/* Shift counts for extracting the individual AccessRight settings. */
#define DF_GET_READ_ONLY                    12
#define DF_GET_WRITE_ONLY                   8
#define DF_GET_READ_WRITE                   4

/* AccessRights masks. */
#define DF_READ_ONLY                        0xF000
#define DF_WRITE_ONLY                       0x0F00
#define DF_READ_WRITE                       0x00F0
#define DF_CHANGE_ACCESS_RIGHTS             0x000F

/*
 * DF_VERSION_INFO
 * ---------------
 * Structure for returning the information supplied by the GetVersion command.
 */
typedef struct
{
BYTE    bHwVendorID;
BYTE    bHwType;
BYTE    bHwSubType;
BYTE    bHwMajorVersion;
BYTE    bHwMinorVersion;
BYTE    bHwStorageSize;
BYTE    bHwProtocol;
BYTE    bSwVendorID;
BYTE    bSwType;
BYTE    bSwSubType;
BYTE    bSwMajorVersion;
BYTE    bSwMinorVersion;
BYTE    bSwStorageSize;
BYTE    bSwProtocol;
BYTE    abUid[7];
BYTE    abBatchNo[5];
BYTE    bProductionCW;
BYTE    bProductionYear;
}
DF_VERSION_INFO;

/*
 * DF_ADDITIONAL_FILE_SETTINGS
 * ---------------------------
 * Union for returning the information supplied by the GetFileSettings command.
 * Use stDataFileSettings for Standard Data Files and Backup Data Files.
 * Use stValueFileSettings for Value Files.
 * Use stRecordFileSettings for Linear Record Files and Cyclic Record Files.
 */
typedef union
{
struct
{
DWORD   eFileSize;
}
stDataFileSettings;

struct
{
SDWORD  lLowerLimit;
SDWORD  lUpperLimit;
DWORD   eLimitedCredit;
BYTE    bLimitedCreditEnabled;
}
stValueFileSettings;

struct
{
DWORD   eRecordSize;
DWORD   eMaxNRecords;
DWORD   eCurrNRecords;
}
stRecordFileSettings;

}
DF_ADDITIONAL_FILE_SETTINGS;

/* Data item lengths for the GetFileSettings command */
#define	DF_FILE_SETTINGS_COMMON_LENGTH	4
#define	DF_DATA_FILE_SETTINGS_LENGTH	  3
#define	DF_VALUE_FILE_SETTINGS_LENGTH	 13
#define	DF_RECORD_FILE_SETTINGS_LENGTH  9







#define	DF_APPLICATION_ID_SIZE			3
#define	DF_MAX_INFO_FRAME_SIZE			60

#ifdef __cplusplus
extern  "C"
{
#endif

/*
 * DesFire exported functions
 * --------------------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Authenticate(BYTE bKeyNumber,
                                                                   const BYTE * pbAccessKey);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ChangeKeySettings(BYTE key_settings);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetKeySettings(BYTE * key_settings,
                                                                     BYTE * key_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ChangeKey(BYTE key_number,
                                                                const BYTE * new_key,
                                                                const BYTE * old_key);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetKeyVersion(BYTE bKeyNumber,
                                                                    BYTE * pbKeyVersion);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_FormatPICC(void);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateApplication(DWORD aid,
                                                                        BYTE key_settings,
                                                                        BYTE keys_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_DeleteApplication(DWORD aid);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetApplicationIDs(BYTE aid_max_count,
                                                                        DWORD * aid_list,
                                                                        BYTE * aid_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_SelectApplication(DWORD aid);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetVersion(DF_VERSION_INFO * pVersionInfo);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetFileIDs(BYTE fid_max_count,
                                                                 BYTE * fid_list,
                                                                 BYTE * fid_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetFileSettings(BYTE file_id,
                                                                      BYTE * file_type,
                                                                      BYTE * comm_mode,
                                                                      WORD * access_rights,
                                                                      DF_ADDITIONAL_FILE_SETTINGS * additionnal_settings);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ChangeFileSettings(BYTE file_id,
                                                                         BYTE comm_mode,
                                                                         WORD access_rights);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateStdDataFile(BYTE file_id,
                                                                        BYTE comm_mode,
                                                                        WORD access_rights,
                                                                        DWORD file_size);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateBackupDataFile(BYTE file_id,
                                                                           BYTE comm_mode,
                                                                           WORD access_rights,
                                                                           DWORD file_size);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateValueFile(BYTE file_id,
                                                                      BYTE comm_mode,
                                                                      WORD access_rights,
                                                                      SDWORD lower_limit,
                                                                      SDWORD upper_limit,
                                                                      SDWORD initial_value,
                                                                      BYTE limited_credit_enabled);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateLinearRecordFile(BYTE file_id,
                                                                             BYTE comm_mode,
                                                                             WORD access_rights,
                                                                             DWORD record_size,
                                                                             DWORD max_records);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateCyclicRecordFile(BYTE file_id,
                                                                             BYTE comm_mode,
                                                                             WORD access_rights,
                                                                             DWORD record_size,
                                                                             DWORD max_records);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_DeleteFile(BYTE file_id);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadData(BYTE file_id,
                                                               BYTE comm_mode,
                                                               DWORD from_offset,
                                                               DWORD max_count,
                                                               BYTE * data,
                                                               DWORD * done_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadData2(BYTE file_id,
                                                                DWORD from_offset,
                                                                DWORD max_count,
                                                                BYTE * data,
                                                                DWORD * done_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteData(BYTE file_id,
                                                                BYTE comm_mode,
                                                                DWORD from_offset,
                                                                DWORD size,
                                                                BYTE * data);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteData2(BYTE file_id,
                                                                 DWORD from_offset,
                                                                 DWORD size,
                                                                 BYTE * data);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetValue(BYTE file_id,
                                                               BYTE comm_mode,
                                                               SDWORD * value);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetValue2(BYTE file_id,
                                                                SDWORD * value);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadRecords(BYTE file_id,
                                                                  BYTE comm_mode,
                                                                  DWORD from_record,
                                                                  DWORD max_record_count,
                                                                  DWORD record_size,
                                                                  BYTE * data,
                                                                  DWORD * record_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadRecords2(BYTE file_id,
                                                                   DWORD from_record,
                                                                   DWORD max_record_count,
                                                                   DWORD record_size,
                                                                   BYTE * data,
                                                                   DWORD * record_count);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteRecord(BYTE file_id,
                                                                  BYTE comm_mode,
                                                                  DWORD from_offset,
                                                                  DWORD size,
                                                                  BYTE * data);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteRecord2(BYTE file_id,
                                                                   DWORD from_offset,
                                                                   DWORD size,
                                                                   BYTE * data);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_LimitedCredit(BYTE file_id,
                                                                    BYTE comm_mode,
                                                                    SDWORD amount);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_LimitedCredit2(BYTE file_id,
                                                                     SDWORD amount);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Credit(BYTE file_id,
                                                             BYTE comm_mode,
                                                             SDWORD amount);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Credit2(BYTE file_id,
                                                              SDWORD amount);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Debit(BYTE file_id,
                                                            BYTE comm_mode,
                                                            SDWORD amount);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Debit2(BYTE file_id,
                                                             SDWORD amount);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ClearRecordFile(BYTE file_id);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CommitTransaction(void);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_AbortTransaction(void);

SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_SelectCid(BYTE cid);

/*
 * DesFire helpers
 * ---------------
 */
SPROX_DESFIRE_LIB const TCHAR *SPROX_DESFIRE_API DesFire_GetErrorMessage(SWORD status);


/*
 * New for dummy .NET languages
 * ----------------------------
 * Caller must provide sizeof(additionnal_settings_array) > 13 !!!
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ExplainDataFileSettings(BYTE * additionnal_settings_array,
                                                                              DWORD * eFileSize);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ExplainValueFileSettings(BYTE * additionnal_settings_array,
                                                                               SDWORD * lLowerLimit,
                                                                               SDWORD * lUpperLimit,
                                                                               DWORD * eLimitedCredit,
                                                                               BYTE * bLimitedCreditEnabled);
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ExplainRecordFileSettings(BYTE * additionnal_settings_array,
                                                                                DWORD * eRecordSize,
                                                                                DWORD * eMaxNRecords,
                                                                                DWORD * eCurrNRecords);

#ifdef __cplusplus
}
#endif


#endif
