/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  SpringProx API
  --------------

  Copyright (c) 2000-2006 Pro-Active SAS - www.pro-active.fr

  config_win32.c
  --------------
  Where to store configuration settings under Win32

  revision :
  ----------

	JDA 09/04/2004 : created
	JDA 26/08/2004 : added device_specific_info stuff
	JDA 22/09/2006 : added the ability to put DeviceName in registry

*/
#include "sprox_api_i.h"

#ifdef WIN32

extern BOOL skip_vcp_check;

#define HKEY_SPRINGPROX       _T("Software\\Pro-Active\\SpringProx")

static BOOL GetDefaultDeviceFromRegistry(SPROX_CTX_ST *sprox_ctx);
static BOOL GetDefaultDeviceFromHardware(SPROX_CTX_ST *sprox_ctx);

BOOL GetDefaultDevice(SPROX_CTX_ST *sprox_ctx)
{
#ifdef UNDER_CE 
  if (GetDefaultDeviceFromHardware(sprox_ctx)) return TRUE;
#endif
  if (GetDefaultDeviceFromRegistry(sprox_ctx)) return TRUE;
  return FALSE;
}

static BOOL GetDefaultDeviceFromRegistry(SPROX_CTX_ST *sprox_ctx)
{
  HKEY    hKey;
  DWORD   size;

  if (sprox_ctx == NULL)
    return FALSE;
    
  sprox_ctx->com_type = 0;

  /* Open the HKCU\Sofware\Pro-Active\SpringProx key in registry */
  if (RegOpenKeyEx(HKEY_CURRENT_USER, HKEY_SPRINGPROX, 0, KEY_READ|KEY_EXECUTE, &hKey) == ERROR_SUCCESS)
  {
    /* Get subkeys count */
    size = sizeof(sprox_ctx->com_name);
    if (RegQueryValueEx(hKey, _T("DeviceName"), NULL, NULL, (LPBYTE) sprox_ctx->com_name, &size) == ERROR_SUCCESS)
    {
      /* Found ! */
      RegCloseKey(hKey);
      return TRUE;
    }
    RegCloseKey(hKey);
  }

  /* Open the HKLM\Sofware\Pro-Active\SpringProx key in registry */
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, HKEY_SPRINGPROX, 0, KEY_READ|KEY_EXECUTE, &hKey) == ERROR_SUCCESS)
  {
    /* Get subkeys count */
    size = sizeof(sprox_ctx->com_name);
    if (RegQueryValueEx(hKey, _T("DeviceName"), NULL, NULL, (LPBYTE) sprox_ctx->com_name, &size) == ERROR_SUCCESS)
    {
      /* Found ! */
      RegCloseKey(hKey);
      return TRUE;
    }
    RegCloseKey(hKey);
  }

  return FALSE;
}


static BOOL GetDefaultDeviceFromHardware(SPROX_CTX_ST *sprox_ctx)
{
#ifdef _SMA1020_

  /* TRICUBES specific : device is physically bound to COM7 */
  if (sprox_ctx == NULL)
    return FALSE;

  sprox_ctx->com_type = 0;
  _tcscpy(sprox_ctx->com_name, _T("COM7:"));
  return TRUE;

#else

  DWORD   i;

  HKEY    hKey, hSubKey;
  DWORD   size;
  TCHAR   shortBuffer[32];
  TCHAR   longBuffer[256];

  if (sprox_ctx == NULL)
    return FALSE;

  sprox_ctx->com_type = 0;

  /* Open the HKLM\Drivers\Active key in registry */
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("Drivers\\Active"), 0, KEY_READ|KEY_EXECUTE, &hKey) == ERROR_SUCCESS)
  {
    /* Get subkeys count */
    if (RegQueryInfoKey(hKey, NULL, NULL, NULL, &i, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
    {
      /* Browse subkeys in reverse order (pluggable cards are not at the beginning of the list !) */
      while (i)
      {
        i--;
        /* Select the subkey */
        size = sizeof(shortBuffer);
        if (RegEnumKeyEx(hKey, i, shortBuffer, &size, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
        {
          /* Open the subkey */
          shortBuffer[sizeof(shortBuffer) - 1] = '\0';
          _stprintf(longBuffer, _T("Drivers\\Active\\%s"), shortBuffer);
          if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, longBuffer, 0, 0, &hSubKey) == ERROR_SUCCESS)
          {
            /* Read the PnPID value */
            size = sizeof(longBuffer);
            if (RegQueryValueEx(hSubKey, _T("PnPID"), NULL, NULL, (LPBYTE) longBuffer, &size) == ERROR_SUCCESS)
            {
              /* Compare the PnPID to requested one */
              if (_tcsncmp(longBuffer, SPRINGPROX_PNPID, _tcsclen(SPRINGPROX_PNPID)) == 0)
              {
                /* This is the good PnPID, now get the serial com in the "Name" value */
                size = sizeof(sprox_ctx->com_name);
                if (RegQueryValueEx(hSubKey, _T("Name"), NULL, NULL, (LPBYTE) sprox_ctx->com_name, &size) == ERROR_SUCCESS)
                {
                  /* Found ! */
                  RegCloseKey(hSubKey);
                  RegCloseKey(hKey);
                  /* OK, note this is a CF reader */
                  sprox_ctx->com_type = DEVICE_IS_CF | DEVICE_POWER_AUTO;
                  return TRUE;
                }
              }
            }

            /* New 1.20 (for SpringXXXX-SD) : lookup for Arasan device */
            size = sizeof(longBuffer);
            if (RegQueryValueEx(hSubKey, _T("Key"), NULL, NULL, (LPBYTE) longBuffer, &size) == ERROR_SUCCESS)
            {
              if (_tcsncmp(longBuffer, ARASAN_DEVICE_KEY_STR, _tcsclen(ARASAN_DEVICE_KEY_STR)) == 0)
              {
                /* This is the good PnPID, now get the serial com in the "Name" value */
                size = sizeof(sprox_ctx->com_name);
                if (RegQueryValueEx(hSubKey, _T("Name"), NULL, NULL, (LPBYTE) sprox_ctx->com_name, &size) == ERROR_SUCCESS)
                {
                  /* Found ! */
                  RegCloseKey(hSubKey);
                  RegCloseKey(hKey);
                  /* OK, note this is an SDIO reader */
                  sprox_ctx->com_type = DEVICE_IS_SDIO | DEVICE_POWER_AUTO;
                  return TRUE;
                }
              }
            }

            /* Close the subkey */
            RegCloseKey(hSubKey);
          }
        }
      }
    }
  }
  /* Close the key */
  RegCloseKey(hKey);

  /* Not found */
  return FALSE;
#endif
}


/*
 * Retrieve (and apply) debug level
 * --------------------------------
 */
void LoadSettings(void)
{
  HKEY    hKey;
  DWORD   size;
  TCHAR   buffer[256];
  BYTE    loop;
  BYTE    level = 0;
  
  for (loop=0; loop<=1; loop++)
  {
    if (loop == 0)
    {
       /* First step : try to locate debug config under HKCU\Sofware\Pro-Active\SpringProx */
      if (!RegOpenKeyEx(HKEY_CURRENT_USER, HKEY_SPRINGPROX, 0, KEY_READ|KEY_EXECUTE, &hKey) == ERROR_SUCCESS)
        continue;
    } else
    {
      /* Second step : do the same under HKLM\Sofware\Pro-Active\SpringProx */
      if (!RegOpenKeyEx(HKEY_LOCAL_MACHINE, HKEY_SPRINGPROX, 0, KEY_READ|KEY_EXECUTE, &hKey) == ERROR_SUCCESS)
        continue;
    }
    
    size = sizeof(buffer);
    if (RegQueryValueEx(hKey, _T("DebugLevel"), NULL, NULL, (LPBYTE) buffer, &size) == ERROR_SUCCESS)
      level = (BYTE) _ttoi(buffer);

    size = sizeof(buffer);        
    if (RegQueryValueEx(hKey, _T("DebugFile"), NULL, NULL, (LPBYTE) buffer, &size) == ERROR_SUCCESS)   
      SPROX_TraceSetFile(buffer);

    size = sizeof(buffer);        
    if (RegQueryValueEx(hKey, _T("SkipVCP"), NULL, NULL, (LPBYTE) buffer, &size) == ERROR_SUCCESS)   
      if (_ttoi(buffer))
        skip_vcp_check = TRUE;

    RegCloseKey(hKey);
  }
  
  if (level)
  {
    SPROX_TraceSetLevel(level);
    if (level >= 6)
    {
      MessageBox(0, _T("The DLL is currently running with highest debug level. Be careful that debug file will become rapidly very very big !"), _T("SpringProx API"), 0);      
    }
  }
}

#ifndef UNDER_CE

static BOOL GetCommName(const TCHAR *device, TCHAR *name, size_t namelen)
{
  HKEY hkey;
  DWORD idx, nl, vl;
  TCHAR n[256], v[256];

  if (device == NULL)
    return FALSE;

  /* Ouverture de la cle qui contient la liste des ports series */
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("HARDWARE\\DEVICEMAP\\SERIALCOMM"), 0, KEY_READ, &hkey) != ERROR_SUCCESS)
    return FALSE;

  idx = 0;
  for (idx=0; ; idx++)
  {
    nl = sizeof(n) / sizeof(TCHAR);
    vl = sizeof(v) / sizeof(TCHAR);
    if (RegEnumValue(hkey, idx, n, &nl, NULL, NULL, v, &vl) != ERROR_SUCCESS)
      break;

    if (!_tcscmp(v, device))
    {
      _tcsncpy(name, n, namelen);
      name[namelen-1] = '\0';
      RegCloseKey(hkey);
      return TRUE;
    }
  }

  RegCloseKey(hkey);
  return FALSE;
}

BOOL IsCommUSB_FTDI(const TCHAR *device)
{
  TCHAR name[64];

  if (!GetCommName(device, name, sizeof(name)/sizeof(TCHAR)))
    return FALSE;

  if (_tcsstr(name, _T("VCP")))
    return TRUE;

  return FALSE;
}

BOOL IsCommUSB_CDC(const TCHAR *device)
{
  TCHAR name[64];

  if (!GetCommName(device, name, sizeof(name)/sizeof(TCHAR)))
    return FALSE;

  if (_tcsstr(name, _T("USBSER")))
    return TRUE;

  return FALSE;
}

#ifndef SPROX_API_NO_USB
/**f* SpringProxAPI/SPROX_EnumUSBDevices
 *
 * NAME
 *   SPROX_EnumUSBDevices
 *
 * DESCRIPTION
 *   Used to enumerate available USB devices that are supported by the SpringProx library
 *   (WIN32 with USB support only)
 *
 * INPUTS
 *   DWORD idx             : index of the device to query
 *   TCHAR device[64]      : buffer to receive the device identifier (mandatory)
 *   TCHAR description[64] : buffer to receive the device descriptor (optional)
 *
 * RETURNS
 *   MI_OK                 : success, device (and description) have been provided
 *   MI_WRONG_VALUE        : invalid value for idx, end of the loop
 *   MI_SER_ACCESS_ERR     : error during query
 *
 **/
SPROX_API_FUNC_EX(EnumUSBDevices) (DWORD idx, TCHAR device[64], TCHAR description[64])
{
  FT_STATUS    ftStatus;
  DWORD        count;
  char         buffer[64];
  unsigned int i;
  
  if (device == NULL)
    return MI_LIB_CALL_ERROR;

  ftStatus = FT_ListDevices(&count, NULL, FT_LIST_NUMBER_ONLY);
  if (ftStatus == FT_OK)
  {
    /* FT_ListDevices OK, number of devices connected is in numDevs */
    if (idx >= count)
      return MI_WRONG_VALUE;
  } else
  {
    /* FT_ListDevices failed */
    return MI_SER_ACCESS_ERR;
  }

  ftStatus = FT_ListDevices((PVOID) idx, buffer, FT_LIST_BY_INDEX|FT_OPEN_BY_SERIAL_NUMBER);
  if (ftStatus == FT_OK)
  {
    /* FT_ListDevices OK, serial number is in Buffer */
    device[0] = 'U'; device[1] = 'S'; device[2] = 'B'; device[3] = ':';
    for (i=0; i<=strlen(buffer); i++)
    {
      if (i>=60) break;
      device[4+i] = buffer[i];
    }
    device[63] = '\0';
  } else
  {
    /* FT_ListDevices failed */
    return MI_SER_ACCESS_ERR;
  }
  
  if (description != NULL)
  {
    ftStatus = FT_ListDevices((PVOID) idx, buffer, FT_LIST_BY_INDEX|FT_OPEN_BY_DESCRIPTION);
    if (ftStatus == FT_OK)
    {
      /* FT_ListDevices OK, description is in Buffer */
      for (i=0; i<=strlen(buffer); i++)
      {
        if (i>=64) break;
        description[i] = buffer[i];
      }
      description[63] = '\0';
    } else
    {
      /* FT_ListDevices failed */
      return MI_SER_ACCESS_ERR;
    }
  }

  return MI_OK;
}
#endif
#endif

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
  __try
  {
    switch (ul_reason_for_call)
    {
      case DLL_PROCESS_ATTACH:
  			LoadSettings();
#ifndef SPROX_API_REENTRANT
        memset(sprox_ctx_glob, 0, sizeof(SPROX_CTX_ST));
        sprox_ctx_glob->com_handle = INVALID_HANDLE_VALUE;
#endif
        break;
      case DLL_THREAD_ATTACH:
        break;
      case DLL_THREAD_DETACH:
        break;
      case DLL_PROCESS_DETACH:
#ifndef SPROX_API_REENTRANT
        SPROX_ReaderClose();
#endif
        break;
    }
    return TRUE;
  }
  __except(1)
  {
    return FALSE;
  }
}

#endif
