/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  SpringProx API
  --------------

  Copyright (c) 2000-2006 Pro-Active SAS - www.pro-active.fr

  serial_usb_win32.c
  ------------------
  serial communication on top of USB, under Windows

  revision :
  ----------

  JDA 30/07/2003 : creation

*/
#include "sprox_api_i.h"

#ifdef WIN32
#ifndef UNDER_CE
#ifndef SPROX_API_NO_USB

static HINSTANCE hModule = NULL;
static BOOL USB_ENABLED = FALSE;

typedef BOOL(WINAPI T_FT_W32_ReadFile) (FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD nBufferSize, LPDWORD lpBytesReturned, LPOVERLAPPED lpOverlapped);
typedef BOOL(WINAPI T_FT_W32_WriteFile) (FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD nBufferSize, LPDWORD lpBytesWritten, LPOVERLAPPED lpOverlapped);
typedef FT_HANDLE(WINAPI T_FT_W32_CreateFile) (LPCSTR lpszName, DWORD dwAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreate,
                                               DWORD dwAttrsAndFlags, HANDLE hTemplate);
typedef BOOL(WINAPI T_FT_W32_CloseHandle) (FT_HANDLE ftHandle);
typedef BOOL(WINAPI T_FT_W32_GetCommState) (FT_HANDLE ftHandle, LPFTDCB lpftDcb);
typedef BOOL(WINAPI T_FT_W32_SetCommState) (FT_HANDLE ftHandle, LPFTDCB lpftDcb);
typedef BOOL(WINAPI T_FT_W32_GetCommTimeouts) (FT_HANDLE ftHandle, FTTIMEOUTS * pTimeouts);
typedef BOOL(WINAPI T_FT_W32_SetCommTimeouts) (FT_HANDLE ftHandle, FTTIMEOUTS * pTimeouts);
typedef BOOL(WINAPI T_FT_W32_ClearCommError) (FT_HANDLE ftHandle, LPDWORD lpdwErrors, LPFTCOMSTAT lpftComstat);
typedef BOOL(WINAPI T_FT_W32_PurgeComm) (FT_HANDLE ftHandle, DWORD dwMask);

typedef FT_STATUS(WINAPI T_FT_ResetDevice) (FT_HANDLE ftHandle);
typedef FT_STATUS(WINAPI T_FT_SetUSBParameters) (FT_HANDLE ftHandle, ULONG ulInTransferSize, ULONG ulOutTransferSize);
typedef FT_STATUS(WINAPI T_FT_SetLatencyTimer) (FT_HANDLE ftHandle, UCHAR ucLatency);
typedef FT_STATUS(WINAPI T_FT_ListDevices) (PVOID pArg1, PVOID pArg2, DWORD Flags);
typedef DWORD(WINAPI T_FT_W32_GetLastError) (FT_HANDLE ftHandle);

T_FT_W32_ReadFile *p_FT_W32_ReadFile;
T_FT_W32_WriteFile *p_FT_W32_WriteFile;
T_FT_W32_CreateFile *p_FT_W32_CreateFile;
T_FT_W32_CloseHandle *p_FT_W32_CloseHandle;
T_FT_W32_GetCommState *p_FT_W32_GetCommState;
T_FT_W32_SetCommState *p_FT_W32_SetCommState;
T_FT_W32_GetCommTimeouts *p_FT_W32_GetCommTimeouts;
T_FT_W32_SetCommTimeouts *p_FT_W32_SetCommTimeouts;
T_FT_W32_ClearCommError *p_FT_W32_ClearCommError;
T_FT_W32_PurgeComm *p_FT_W32_PurgeComm;
T_FT_W32_GetLastError *p_FT_W32_GetLastError;

T_FT_ResetDevice *p_FT_ResetDevice;
T_FT_SetUSBParameters *p_FT_SetUSBParameters;
T_FT_SetLatencyTimer *p_FT_SetLatencyTimer;
T_FT_ListDevices *p_FT_ListDevices;

BOOL FT_W32_Initialize(void)
{
  if (USB_ENABLED)
    return TRUE;

  USB_ENABLED = FALSE;
#ifdef UNICODE
  hModule = LoadLibraryW(L"ftd2xx.dll");
#else
  hModule = LoadLibraryA("ftd2xx.dll");
#endif

  if (hModule == NULL)
    return FALSE;
  USB_ENABLED = TRUE;

  p_FT_W32_CreateFile = (T_FT_W32_CreateFile *) GetProcAddress(hModule, "FT_W32_CreateFile");
  if (p_FT_W32_CreateFile == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_CloseHandle = (T_FT_W32_CloseHandle *) GetProcAddress(hModule, "FT_W32_CloseHandle");
  if (p_FT_W32_CloseHandle == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_ReadFile = (T_FT_W32_ReadFile *) GetProcAddress(hModule, "FT_W32_ReadFile");
  if (p_FT_W32_ReadFile == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_WriteFile = (T_FT_W32_WriteFile *) GetProcAddress(hModule, "FT_W32_WriteFile");
  if (p_FT_W32_WriteFile == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_GetCommState = (T_FT_W32_GetCommState *) GetProcAddress(hModule, "FT_W32_GetCommState");
  if (p_FT_W32_GetCommState == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_SetCommState = (T_FT_W32_SetCommState *) GetProcAddress(hModule, "FT_W32_SetCommState");
  if (p_FT_W32_SetCommState == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_GetCommTimeouts = (T_FT_W32_GetCommTimeouts *) GetProcAddress(hModule, "FT_W32_GetCommTimeouts");
  if (p_FT_W32_GetCommTimeouts == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_SetCommTimeouts = (T_FT_W32_SetCommTimeouts *) GetProcAddress(hModule, "FT_W32_SetCommTimeouts");
  if (p_FT_W32_SetCommTimeouts == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_ClearCommError = (T_FT_W32_ClearCommError *) GetProcAddress(hModule, "FT_W32_ClearCommError");
  if (p_FT_W32_ClearCommError == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_PurgeComm = (T_FT_W32_PurgeComm *) GetProcAddress(hModule, "FT_W32_PurgeComm");
  if (p_FT_W32_PurgeComm == NULL)
    USB_ENABLED = FALSE;

  p_FT_W32_GetLastError = (T_FT_W32_GetLastError *) GetProcAddress(hModule, "FT_W32_GetLastError");
  if (p_FT_W32_GetLastError == NULL)
    USB_ENABLED = FALSE;

  p_FT_ResetDevice = (T_FT_ResetDevice *) GetProcAddress(hModule, "FT_ResetDevice");
  if (p_FT_ResetDevice == NULL)
    USB_ENABLED = FALSE;

  p_FT_SetUSBParameters = (T_FT_SetUSBParameters *) GetProcAddress(hModule, "FT_SetUSBParameters");
  if (p_FT_SetUSBParameters == NULL)
    USB_ENABLED = FALSE;

  p_FT_SetLatencyTimer = (T_FT_SetLatencyTimer *) GetProcAddress(hModule, "FT_SetLatencyTimer");
  if (p_FT_SetLatencyTimer == NULL)
    USB_ENABLED = FALSE;

  p_FT_ListDevices = (T_FT_ListDevices *) GetProcAddress(hModule, "FT_ListDevices");
  if (p_FT_ListDevices == NULL)
    USB_ENABLED = FALSE;

  if (!USB_ENABLED)
  {
    FreeLibrary(hModule);
    hModule = NULL;
  } else
  {

  }
  return USB_ENABLED;
}

void FT_W32_Finalize(void)
{
  if (hModule != NULL)
  {
    FreeLibrary(hModule);
    hModule = NULL;
  }
  USB_ENABLED = FALSE;
}

FTD2XX_API FT_HANDLE WINAPI FT_W32_CreateFile(LPCSTR lpszName, DWORD dwAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreate,
                                              DWORD dwAttrsAndFlags, HANDLE hTemplate)
{
  __try
  {
    if (!FT_W32_Initialize())
      return NULL;
    return p_FT_W32_CreateFile(lpszName, dwAccess, dwShareMode, lpSecurityAttributes, dwCreate, dwAttrsAndFlags, hTemplate);
  }
  __except(1)
  {
    return NULL;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_ReadFile(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD nBufferSize, LPDWORD lpBytesReturned, LPOVERLAPPED lpOverlapped)
{
  if (!USB_ENABLED)
    return FALSE;

  __try
  {
    BOOL    rc;
    rc = p_FT_W32_ReadFile(ftHandle, lpBuffer, nBufferSize, lpBytesReturned, lpOverlapped);
    if (!rc)
    {
      if (FT_W32_GetLastError(ftHandle) == FT_IO_ERROR)
        FT_W32_Finalize();
    }
    return rc;
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_WriteFile(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD nBufferSize, LPDWORD lpBytesWritten, LPOVERLAPPED lpOverlapped)
{
  if (!USB_ENABLED)
    return FALSE;

  __try
  {
    BOOL    rc;
    rc = p_FT_W32_WriteFile(ftHandle, lpBuffer, nBufferSize, lpBytesWritten, lpOverlapped);
    if (!rc)
    {
      if (FT_W32_GetLastError(ftHandle) == FT_IO_ERROR)
        FT_W32_Finalize();
    }
    return rc;
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_CloseHandle(FT_HANDLE ftHandle)
{
  __try
  {
    BOOL    rc = TRUE;
    if (USB_ENABLED)
    {
      rc = p_FT_W32_CloseHandle(ftHandle);
      FT_W32_Finalize();
    }
    return rc;
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_GetCommState(FT_HANDLE ftHandle, LPFTDCB lpftDcb)
{
  if (!USB_ENABLED)
    return FALSE;
  __try
  {
    return p_FT_W32_GetCommState(ftHandle, lpftDcb);
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_SetCommState(FT_HANDLE ftHandle, LPFTDCB lpftDcb)
{
  if (!USB_ENABLED)
    return FALSE;
  __try
  {
    return p_FT_W32_SetCommState(ftHandle, lpftDcb);
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_GetCommTimeouts(FT_HANDLE ftHandle, FTTIMEOUTS * pTimeouts)
{
  if (!USB_ENABLED)
    return FALSE;
  __try
  {
    return p_FT_W32_GetCommTimeouts(ftHandle, pTimeouts);
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_SetCommTimeouts(FT_HANDLE ftHandle, FTTIMEOUTS * pTimeouts)
{
  if (!USB_ENABLED)
    return FALSE;
  __try
  {
    return p_FT_W32_SetCommTimeouts(ftHandle, pTimeouts);
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_ClearCommError(FT_HANDLE ftHandle, PDWORD lpdwErrors, LPFTCOMSTAT lpftComstat)
{
  __try
  {
    DWORD   arg1;
    FTCOMSTAT arg2;
    if (!USB_ENABLED)
      return FALSE;
    if (lpdwErrors == NULL)
      lpdwErrors = &arg1;
    if (lpftComstat == NULL)
      lpftComstat = &arg2;
    return p_FT_W32_ClearCommError(ftHandle, lpdwErrors, lpftComstat);
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API BOOL WINAPI FT_W32_PurgeComm(FT_HANDLE ftHandle, DWORD dwMask)
{
  if (!USB_ENABLED)
    return FALSE;
  __try
  {
    return p_FT_W32_PurgeComm(ftHandle, dwMask);
  }
  __except(1)
  {
    return FALSE;
  }
}

FTD2XX_API DWORD WINAPI FT_W32_GetLastError(FT_HANDLE ftHandle)
{
  if (!USB_ENABLED)
    return 0;
  __try
  {
    return p_FT_W32_GetLastError(ftHandle);
  }
  __except(1)
  {
    return 0;
  }
}

FTD2XX_API FT_STATUS WINAPI FT_ListDevices(PVOID pArg1, PVOID pArg2, DWORD Flags)
{
  __try
  {
    if (!FT_W32_Initialize())
      return FT_INVALID_HANDLE;
    return p_FT_ListDevices(pArg1, pArg2, Flags);
  }
  __except(1)
  {
    return FT_OTHER_ERROR;
  }
}

FTD2XX_API FT_STATUS WINAPI FT_SetLatencyTimer(FT_HANDLE ftHandle, UCHAR ucLatency)
{
  if (!USB_ENABLED)
    return FT_INVALID_HANDLE;
  __try
  {
    return p_FT_SetLatencyTimer(ftHandle, ucLatency);
  }
  __except(1)
  {
    return FT_OTHER_ERROR;
  }
}

FTD2XX_API FT_STATUS WINAPI FT_SetUSBParameters(FT_HANDLE ftHandle, ULONG ulInTransferSize, ULONG ulOutTransferSize)
{
  if (!USB_ENABLED)
    return FT_INVALID_HANDLE;
  __try
  {
    return p_FT_SetUSBParameters(ftHandle, ulInTransferSize, ulOutTransferSize);
  }
  __except(1)
  {
    return FT_OTHER_ERROR;
  }
}

FTD2XX_API FT_STATUS WINAPI FT_ResetDevice(FT_HANDLE ftHandle)
{
  if (!USB_ENABLED)
    return FT_INVALID_HANDLE;
  __try
  {
    return p_FT_ResetDevice(ftHandle);
  }
  __except(1)
  {
    return FT_OTHER_ERROR;
  }
}



#endif
#endif
#endif
