/**h* SpringProxAPI/Wait
 *
 * NAME
 *   SpringProxAPI :: Automatic card discovery functions
 *
 * DESCRIPTION
 *   Those functions work only for readers having version >= 1.44
 *
 * PORTABILITY
 *   Win32 only for the multi-threaded and event driven part
 *
 **/

/*

  SpringProx API
  --------------

  Copyright (c) 2000-2008 Pro-Active SAS - www.pro-active.fr

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  History
  -------
  
  JDA 28/02/2008 : created
  
*/

#include "sprox_api_i.h"

/**f* SpringProxAPI/SPROX_Find
 *
 * NAME
 *   SPROX_Find
 *
 * DESCRIPTION
 *   Automatic discovery of any contactless card in the field
 *
 * NOTES
 *   This function is only avalaible for readers having version >= 1.44
 *
 * INPUTS
 *   WORD want_protos   : bit-map of contactless family to look for
 *   WORD *got_proto    : on exit, found contactless family
 *                        (one bit only is set, or none)
 *   BYTE uid[12]       : Unique IDentifier of the found card
 *   BYTE *uidlen       : on input, size of UID
 *                        on output, actual length of UID
 *
 * RETURNS
 *   MI_OK              : success, one card selected
 *   MI_NOTAGERR        : no card available in the RF field
 *   Other code if internal or communication error has occured.  
 *
 **/
SPROX_API_FUNC_EX(Find) (SPROX_PARAM  WORD want_protos, WORD *got_proto, BYTE uid[12], BYTE *uidlen)
{
  BYTE buffer[64];
  SWORD rc;
  WORD buflen;
  SPROX_PARAM_TO_CTX;
     
  buffer[0] = (BYTE) (want_protos / 0x0100);
  buffer[1] = (BYTE) (want_protos % 0x0100);
  
  buflen = sizeof(buffer);
  rc = SPROX_DLG_FUNC(SPROX_PARAM_P  SPROX_FIND, buffer, 2, buffer, &buflen);
  
  if (rc == MI_OK)
  {
    if (buflen < 2)
    {
      rc = MI_SER_LENGTH_ERR;
    } else
    {
      if (got_proto != NULL)
      {
        *got_proto  = buffer[0];
        *got_proto *= 0x0100;
        *got_proto += buffer[1];
      }
      
      if (uid != NULL)
        memcpy(uid, &buffer[2], buflen - 2);
        
      if (uidlen != NULL)
        *uidlen = buflen - 2;
    }
  }
  
  return rc;
}

#ifdef WIN32

SPROX_API_FUNC_EX(FindStart) (SPROX_PARAM  WORD want_protos, DWORD wait_sec, HANDLE wait_event)
{
  if (wait_event == INVALID_HANDLE_VALUE) return MI_LIB_CALL_ERROR;
  ResetEvent(wait_event);
  

}

SPROX_API_FUNC_EX(FindResult) (SPROX_PARAM  WORD *got_proto, BYTE uid[12], BYTE *uidlen)
{
  
}

SPROX_API_FUNC_EX(FindCancel) (SPROX_PARAM_V)
{
  
}


#endif
