#include "sprox_desfire_i.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef UNDER_CE
#include <time.h>
#endif

/*
 * Protypes
 * --------
 */

/*
 * Function from SpringProx.dll not explictly available...
 */
SPRINGPROX_LIB SWORD SPRINGPROX_API SPROX_TclA_ExchangeDF(BYTE cid, BYTE * send_buffer, WORD send_len, BYTE * recv_buffer, WORD * recv_len);

/*
 * Functions from the desfire_crypto module
 */
void    DesFire_xDES_MAC(BYTE * data, WORD blocks, const BYTE * key, BYTE * result);
void    DesFire_xDES_MAC_b(BYTE * data, DWORD length, const BYTE * key, BYTE * result);
void    DesFire_xDES_Recv(BYTE * data, WORD blocks, const BYTE * key);
void    DesFire_xDES_Send(BYTE * data, WORD blocks, const BYTE * key);

/*
 * Internal module functions
 */
static SWORD DesFireCard_Command(BOOL fast, WORD accept_len, BYTE accept_status);
static SWORD DesFireCard_ReadFile(BYTE read_command, BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD item_count, DWORD item_size, BYTE * data,
                                  DWORD * done_size);
static SWORD DesFireCard_WriteFile(BYTE write_command, BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, BYTE * data);
static SWORD DesFireCard_ModifyValue(BYTE modify_command, BYTE file_id, BYTE comm_mode, SDWORD amount);

/* Flags for DesFireCard_Command */
#define ACCEPT_OPERATION_OK      0x01
#define ACCEPT_ADDITIONAL_FRAME  0x02

/*
 * Internal module globals
 * -----------------------
 */
static BYTE DFC_buffer[4096];
static WORD DFC_buflen = 0;
static BYTE DFC_cid = 0xFF;
static BYTE DFC_auth_key = INVALIDATE_AUTHENTICATION;
static BYTE DFC_session_key[16];

#define INF 0

/*
 * Debug
 */
 
#undef D
#define D(x)

void dump_buffer(const BYTE * buffer, WORD len)
{
  WORD    i;
  for (i = 0; i < len; i++)
  {
    printf("%02X", buffer[i]);
    if (i % 8 == 7)
      printf(" ");
    if (i % 64 == 63)
      printf("\n");
  }
}

/*
 **************************************************************************************************
 *
 *  Exported functions
 *
 *************************************************************************************************
 */

SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_SelectCid(BYTE cid)
{
  DFC_cid = cid;
  return 0;
}

/*
 * Authenticate
 * ------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Authenticate(BYTE bKeyNumber, const BYTE * pbAccessKey)
{
  BOOL    qIsTripleDes = FALSE;
  SWORD   status, w;
//  BYTE abCbcBlock[8];
  BYTE    abRndB[8], abRndA[8];

  /* Each new Authenticate must invalidate the current authentication status. */
  DFC_auth_key = INVALIDATE_AUTHENTICATION;

  /* Check whether a TripleDES key was passed (both key halves are different). */
  if (memcmp(pbAccessKey, pbAccessKey + 8, 8))
  {
    /* If the two key halves are not identical, we are doing TripleDES. */

    /* We have to remember that TripleDES is in effect, because the manner of building
       the session key is different in this case. */
    qIsTripleDes = TRUE;
  }

  /* Create the command string consisting of the command byte and the parameter byte. */
  DFC_buffer[INF + 0] = DF_AUTHENTICATE;
  DFC_buffer[INF + 1] = bKeyNumber;
  DFC_buflen = 2;

  /* Send the command string to the PICC and get its response (1st frame exchange).
     The PICC has to respond with an DF_ADDITIONAL_FRAME status byte. */
  status = DesFireCard_Command(FALSE, 0, ACCEPT_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;

  /* Check the number of bytes received, we expect 9 bytes. */
  if (DFC_buflen != 9)
  {
    /* Error: block with inappropriate number of bytes received from the card. */
    return DFCARD_WRONG_LENGTH;
  }

  /* OK, we received the 8 bytes Ek( RndB ) from the PICC.
     Decipher Ek( RndB ) to get RndB in DFC_buffer.
     Note that the status code has already been extracted from the queue. */
  DesFire_xDES_Recv(&DFC_buffer[INF + 1], 1, pbAccessKey);
  /* Store this RndB (is needed later on for generating the session key). */
  memcpy(abRndB, &DFC_buffer[INF + 1], 8);

  /* Now the PCD has to generate RndA.
     But first seed the random number generator with the current time. */
#ifndef UNDER_CE
  srand((unsigned) time(NULL));
#else
  /* No time on Windows CE */
  srand(0xdead);
#endif
  for (w = 0; w < 8; w++)
    abRndA[w] = (BYTE) rand();

#if 0
  for (w = 0; w < 8; w += 2)
  {
    abRndA[w] = 0x45;
    abRndA[w + 1] = 0x67;
  }
#endif

  D(printf("RndA: "));
  D(dump_buffer(abRndA, 8));
  D(printf("\n"));
  D(printf("RndB: "));
  D(dump_buffer(abRndB, 8));
  D(printf("\n"));

  /* Start the second frame with a status byte indicating to the PICC that the Authenticate
     command is continued. */
  DFC_buffer[INF + 0] = DF_ADDITIONAL_FRAME;

  /* Append RndA and RndB' ( RndB' is generated by rotating RndB one byte to the left )
     after the status byte. */
  memcpy(&DFC_buffer[INF + 1], abRndA, 8);
  memcpy(&DFC_buffer[INF + 9], &abRndB[1], 7);
  DFC_buffer[INF + 16] = abRndB[0]; /* first byte move to last byte */
  DFC_buflen = 17;

  /* Apply the DES send operation to the 16 argument bytes before sending the second frame
     to the PICC ( do not include the status byte in the DES operation ). */
  DesFire_xDES_Send(&DFC_buffer[INF + 1], 2, pbAccessKey);

  /* Send the 2nd frame to the PICC and get its response. */
  status = DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* We should now have Ek( RndA' ) in our buffer.
     RndA' was made from RndA by the PICC by rotating the string one byte to the left.
     Decipher Ek( RndA' ) to get RndA' in DFC_buffer. */
  DesFire_xDES_Recv(&DFC_buffer[INF + 1], 2, pbAccessKey);

  /* Now we have RndA' in our buffer.
     We have to check whether it matches our local copy of RndA.
     If one of the subsequent comparisons fails, we do not trust the PICC and therefore
     abort the authentication procedure ( no session key is generated ). */

  /* First compare the bytes 1 to 7 of RndA with the first 7 bytes in the queue. */
  if (memcmp(&DFC_buffer[INF + 1], abRndA + 1, 7))
    return DFCARD_WRONG_KEY;

  /* Then compare the leftmost byte of RndA with the last byte in the queue. */
  if (DFC_buffer[INF + 8] != abRndA[0])
    return DFCARD_WRONG_KEY;

  /* The actual authentication has succeeded.
     Finally we have to generate the session key from both random numbers RndA and RndB.
     The first half of the session key is the concatenation of RndA[0-3] + RndB[0-3]. */
  memcpy(DFC_session_key, abRndA, 4);
  memcpy(DFC_session_key + 4, abRndB, 4);

  /* If the original key passed through pbAccessKey is a TripleDES key, the session
     key must also be a TripleDES key. */
  if (qIsTripleDes)
  {
    /* For TripleDES generate the second part of the session key.
       This is the concatenation of RndA[4-7] + RndB[4-7]. */
    memcpy(DFC_session_key + 8, abRndA + 4, 4);
    memcpy(DFC_session_key + 12, abRndB + 4, 4);
  } else
  {
    memcpy(DFC_session_key + 8, abRndA, 4);
    memcpy(DFC_session_key + 12, abRndB, 4);
  }

  D(printf("Session Key: "));
  D(dump_buffer(DFC_session_key, 16));
  D(printf("\n"));

  /* Authenticate succeeded, therefore we remember the number of the key which was used
     to obtain the current authentication status. */
  DFC_auth_key = bKeyNumber;

  /* Success. */
  return DF_OPERATION_OK;
}

/*
 * ChangeKeySettings
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ChangeKeySettings(BYTE key_settings)
{
  BYTE    i;

  /* Create the info block containing the command code */
  DFC_buffer[INF + 0] = DF_CHANGE_KEY_SETTINGS;
  DFC_buffer[INF + 1] = key_settings;
  DFC_buflen = 9;               /* 1 byte cmd, 8 bytes of decipher setting */

  /* Append the CRC value corresponding to the key_settings byte */
  /* CRC16 calculation according to 14443A                       */
  SPROX_ComputeIso14443ACrc(&DFC_buffer[INF + 2], &DFC_buffer[INF + 1], 1);

  D(printf("Change Key Settings 1: "));
  D(dump_buffer(&DFC_buffer[INF + 1], 3));
  D(printf("\n"));

  /* 'Encrypt' the bNewKeySettings byte and its CRC bytes padding the DES block with 0 bytes. */
  /* Note: since mifare DESFire always runs in DES ENcryption mode this is actually a DES     */
  /* DEcryption operation.                                                                    */
  for (i = 4; i < 9; i++)
    DFC_buffer[INF + i] = 0;
  DesFire_xDES_Send(&DFC_buffer[INF + 1], 1, DFC_session_key);

  D(printf("Change Key Settings 2: "));
  D(dump_buffer(&DFC_buffer[INF + 2], 8));
  D(printf("\n"));

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetKeySettings
 * --------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetKeySettings(BYTE * key_settings, BYTE * key_count)
{
  SWORD   status;

  /* Create the info block containing the command code */
  DFC_buffer[INF + 0] = DF_GET_KEY_SETTINGS;
  DFC_buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  /* GetKeySettings returns 3 bytes, the status and two bytes of information.        */
  status = DesFireCard_Command(FALSE, 3, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* Return the requested key settings bytes. */
  *key_settings = DFC_buffer[INF + 1];
  *key_count = DFC_buffer[INF + 2];

  /* Success */
  return DF_OPERATION_OK;
}

/*
 * ChangeKey
 * ---------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ChangeKey(BYTE key_number, const BYTE * new_key, const BYTE * old_key)
{
  BYTE    b, buffer[24];

  /* Begin the info block with the command code and the number of the key to be changed. */
  DFC_buffer[INF + 0] = DF_CHANGE_KEY;
  DFC_buffer[INF + 1] = key_number;
  DFC_buflen = 26;

  D(printf("Change Key, new key = "));
  D(dump_buffer(new_key, 16));
  D(printf("\n"));

  if (old_key != NULL)
  {
    /* If a 'previous key' has been passed, format the 24 byte cryptogram according to the     */
    /* three key procedure (new key, previous key, all encrypted with the current session key) */

    /* Take the new key into the buffer. */
    memcpy(buffer, new_key, 16);

    /* XOR the previous key to the new key. */
    for (b = 0; b < 16; b++)
      buffer[b] ^= old_key[b];

    /* Append a first CRC, computed over the XORed key combination. */
    SPROX_ComputeIso14443ACrc(&buffer[16], &buffer[0], 16);

    /* Append a second CRC, computed over the new key only. */
    SPROX_ComputeIso14443ACrc(&buffer[18], (BYTE *) new_key, 16);

    /* Fill the remainder of the 3rd block with zeros (4 bytes). */
    memset(&buffer[20], 0x00, 4);
  } else
  {
    /* If no 'previous key' has been passed, format the 24 byte cryptogram according to the */
    /* two key procedure (new key, encrypted with the current session key).                 */

    /* Take the new key into the buffer. */
    memcpy(buffer, new_key, 16);

    /* Append the CRC computed over the new key. */
    SPROX_ComputeIso14443ACrc(&buffer[16], &buffer[0], 16);

    /* fill the remainder of the 3rd block with zeros (6 bytes). */
    memset(&buffer[18], 0x00, 6);
  }

  /* Append the 24 byte buffer to the command string. */
  memcpy(&DFC_buffer[INF + 2], buffer, 24);

  D(printf("Change Key 1: "));
  D(dump_buffer(&DFC_buffer[INF + 2], 24));
  D(printf("\n"));

  /* 'Encrypt' the 24 key, CRC and padding bytes in CBC mode.                             */
  /* Note: since mifare DESFire always runs in DES ENcryption mode this is actually a DES */
  /* DEcryption operation.                                                                */
  DesFire_xDES_Send(&DFC_buffer[INF + 2], 3, DFC_session_key);

  D(printf("Change Key 2: "));
  D(dump_buffer(&DFC_buffer[INF + 2], 24));
  D(printf("\n"));

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetKeyVersion
 * -------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetKeyVersion(BYTE bKeyNumber, BYTE * pbKeyVersion)
{
  SWORD   status;

  /* Create the info block containing the command code and the key number argument. */
  DFC_buffer[INF + 0] = DF_GET_KEY_VERSION;
  DFC_buffer[INF + 1] = bKeyNumber;
  DFC_buflen = 2;

  /* Communicate the info block to the card and check the operation's return status.   */
  /* The returned info block must contain two bytes, the status code and the requested */
  /* key version byte.                                                                 */
  status = DesFireCard_Command(FALSE, 2, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* Get the key version byte. */
  *pbKeyVersion = DFC_buffer[INF + 1];

  /* Success. */
  return DF_OPERATION_OK;
}

/*
 * FormatPICC
 * ----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_FormatPICC(void)
{
  /* Create the info block containing the command code. */
  DFC_buffer[INF + 0] = DF_FORMAT_PICC;
  DFC_buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateApplication
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateApplication(DWORD aid, BYTE key_settings, BYTE keys_count)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CREATE_APPLICATION;
  DFC_buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  DFC_buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  DFC_buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  DFC_buffer[INF + 4] = key_settings;
  DFC_buffer[INF + 5] = keys_count;
  DFC_buflen = 6;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * DeleteApplication
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_DeleteApplication(DWORD aid)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_DELETE_APPLICATION;
  DFC_buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  DFC_buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  DFC_buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  DFC_buflen = 4;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetApplicationIDs
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetApplicationIDs(BYTE aid_max_count, DWORD * aid_list, BYTE * aid_count)
{
  DWORD   aid;
  BYTE    i;
  WORD    len;
  SWORD   status;

  /* Set the number of applications to zero */
  *aid_count = 0;

  /* create the info block containing the command code */
  DFC_buffer[INF + 0] = DF_GET_APPLICATION_IDS;
  DFC_buflen = 1;

  for (;;)
  {
    status = DesFireCard_Command(FALSE, 0, ACCEPT_ADDITIONAL_FRAME | ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    len = DFC_buflen - 1;       /* substract 1 byte for the received status */

    /* ByteCount must be in multiples of APPLICATION_ID_SIZE bytes
       we check this to proof the format integrity
       if zero bytes have been received this is ok -> means no
       applications existing */
    if (len % APPLICATION_ID_SIZE)
      return DFCARD_WRONG_LENGTH;

    for (i = 0; i < (len / APPLICATION_ID_SIZE); i++)
    {
      /* Extract AID */
      aid = DFC_buffer[INF + 3 + APPLICATION_ID_SIZE * i];
      aid <<= 8;
      aid += DFC_buffer[INF + 2 + APPLICATION_ID_SIZE * i];
      aid <<= 8;
      aid += DFC_buffer[INF + 1 + APPLICATION_ID_SIZE * i];

      /* Copy AID */
      aid_list[*aid_count] = aid;

      /* Next AID */
      *aid_count = *aid_count + 1;
      if (*aid_count >= aid_max_count)
      {
        if (i < ((len / APPLICATION_ID_SIZE) - 1))
        {
          /* Still ApplicationIDs available but we ran out
             of buffer space provided by the caller */
          return DFCARD_OVERFLOW;
        }
      }
    }

    /* Prepare the info block indicating that we want more application IDs */
    if (DFC_buffer[INF + 0] != DF_ADDITIONAL_FRAME)
      return DF_OPERATION_OK;
    DFC_buflen = 1;
  }
}

/*
 * SelectApplication
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_SelectApplication(DWORD aid)
{
  /* Each SelectApplication causes a currently valid authentication state to be lost */
  DFC_auth_key = INVALIDATE_AUTHENTICATION;

  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_SELECT_APPLICATION;
  DFC_buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  DFC_buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  DFC_buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  DFC_buflen = 4;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * GetVersion
 * ----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetVersion(DF_VERSION_INFO * pVersionInfo)
{
  SWORD   status;
  BYTE   *ptr = (BYTE *) pVersionInfo;

  DFC_buffer[INF + 0] = DF_GET_VERSION;
  DFC_buflen = 1;
  status = DesFireCard_Command(FALSE, 8, ACCEPT_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &DFC_buffer[INF + 1], 7);
  ptr += 7;

  DFC_buffer[INF + 0] = DF_ADDITIONAL_FRAME;
  DFC_buflen = 1;
  status = DesFireCard_Command(FALSE, 8, ACCEPT_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &DFC_buffer[INF + 1], 7);
  ptr += 7;

  DFC_buffer[INF + 0] = DF_ADDITIONAL_FRAME;
  DFC_buflen = 1;
  status = DesFireCard_Command(FALSE, 15, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &DFC_buffer[INF + 1], 14);

  return DF_OPERATION_OK;
}

/*
 * GetFileIDs
 * ----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetFileIDs(BYTE fid_max_count, BYTE * fid_list, BYTE * fid_count)
{
  SWORD   status;

  /* reset the number of found FileIDs */
  *fid_count = 0;

  /* create the info block containing the command code */
  DFC_buffer[INF + 0] = DF_GET_FILE_IDS;
  DFC_buflen = 1;

  /* this command terminates within a single frame (no additional frames necessary) */
  status = DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  DFC_buflen--;                 /* subtract 1 byte for the received status of desfire card */
  for (; (*fid_count < fid_max_count) && (DFC_buflen > 0); fid_list++, (*fid_count)++)
  {
    *fid_list = DFC_buffer[INF + 1 + *fid_count];
    DFC_buflen--;
  }

  /* if there are still bytes left in the DFC_buffer   */
  /* we ran out of buffer space provided by the caller */
  if (DFC_buflen != 0)
    return DFCARD_OVERFLOW;

  return DF_OPERATION_OK;
}

/*
 * GetFileSettings
 * ---------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetFileSettings(BYTE file_id, BYTE * file_type, BYTE * comm_mode, WORD * access_rights,
                                                                      DF_ADDITIONAL_FILE_SETTINGS * additionnal_settings)
{
  BYTE    l_file_type;
  SWORD   status;
  DWORD   temp;

  if (file_type == NULL)
    file_type = &l_file_type;

  /* create command code and append FileID parameter */
  DFC_buffer[INF + 0] = DF_GET_FILE_SETTINGS;
  DFC_buffer[INF + 1] = file_id;
  DFC_buflen = 2;

  /* don't do any implicit length check on the response */
  /* this has to be resolved depending on the file type */
  status = DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* there is a fixed section of the response which is common to all file types */
  /* therefore the response must at least be of this fixed length               */
  /* Note the status byte had to be extract!                                    */

  DFC_buflen--;                 /* subtract 1 byte for the received status of desfire card */
  if (DFC_buflen < DF_FILE_SETTINGS_COMMON_LENGTH)
    return DFCARD_WRONG_LENGTH;

  /* extract the file's type */
  *file_type = DFC_buffer[INF + 1];

  /* extract the file's communication mode */
  if (comm_mode != NULL)
    *comm_mode = DFC_buffer[INF + 2];

  /* extract the file's access rights */
  if (access_rights != NULL)
  {
    *access_rights = DFC_buffer[INF + 4];
    *access_rights <<= 8;
    *access_rights += DFC_buffer[INF + 3];
  }

  /* if no pointer to a union for holding the file settings has been passed */
  /* we empty the buffer and return with success                            */
  if (additionnal_settings == NULL)
    return DF_OPERATION_OK;

  /* the rest of the buffer contains file type specific information      */
  /* therefore we have to resolve it according to the received file type */
  DFC_buflen -= DF_FILE_SETTINGS_COMMON_LENGTH; /* subtract for the fixed section */
  switch (*file_type)
  {
    case DF_STANDARD_DATA_FILE:
    case DF_BACKUP_DATA_FILE:

      if (DFC_buflen != DF_DATA_FILE_SETTINGS_LENGTH)
        return DFCARD_WRONG_LENGTH;

      /* extract the file size */
      temp = DFC_buffer[INF + 7];
      temp <<= 8;
      temp += DFC_buffer[INF + 6];
      temp <<= 8;
      temp += DFC_buffer[INF + 5];
      additionnal_settings->stDataFileSettings.eFileSize = temp;
      break;

    case DF_VALUE_FILE:

      if (DFC_buflen != DF_VALUE_FILE_SETTINGS_LENGTH)
        return DFCARD_WRONG_LENGTH;

      /* extract the lower limit */
      temp = DFC_buffer[INF + 8];
      temp <<= 8;
      temp += DFC_buffer[INF + 7];
      temp <<= 8;
      temp += DFC_buffer[INF + 6];
      temp <<= 8;
      temp += DFC_buffer[INF + 5];
      additionnal_settings->stValueFileSettings.lLowerLimit = temp;

      /* extract the upper limit */
      temp = DFC_buffer[INF + 12];
      temp <<= 8;
      temp += DFC_buffer[INF + 11];
      temp <<= 8;
      temp += DFC_buffer[INF + 10];
      temp <<= 8;
      temp += DFC_buffer[INF + 9];
      additionnal_settings->stValueFileSettings.lUpperLimit = temp;

      /* extract the limited credit value; */
      temp = DFC_buffer[INF + 16];
      temp <<= 8;
      temp += DFC_buffer[INF + 15];
      temp <<= 8;
      temp += DFC_buffer[INF + 14];
      temp <<= 8;
      temp += DFC_buffer[INF + 13];
      additionnal_settings->stValueFileSettings.eLimitedCredit = temp;

      /* limited Credit enabled ? */
      additionnal_settings->stValueFileSettings.bLimitedCreditEnabled = DFC_buffer[INF + 17];
      break;

    case DF_LINEAR_RECORD_FILE:
    case DF_CYCLIC_RECORD_FILE:

      if (DFC_buflen != DF_RECORD_FILE_SETTINGS_LENGTH)
        return DFCARD_WRONG_LENGTH;

      /* extract the number of entire records */
      temp = DFC_buffer[INF + 7];
      temp <<= 8;
      temp += DFC_buffer[INF + 6];
      temp <<= 8;
      temp += DFC_buffer[INF + 5];
      additionnal_settings->stRecordFileSettings.eRecordSize = temp;

      /* get the size of one record */
      temp = DFC_buffer[INF + 10];
      temp <<= 8;
      temp += DFC_buffer[INF + 9];
      temp <<= 8;
      temp += DFC_buffer[INF + 8];
      additionnal_settings->stRecordFileSettings.eMaxNRecords = temp;

      /* get the current number of records */
      temp = DFC_buffer[INF + 13];
      temp <<= 8;
      temp += DFC_buffer[INF + 12];
      temp <<= 8;
      temp += DFC_buffer[INF + 11];
      additionnal_settings->stRecordFileSettings.eCurrNRecords = temp;

      break;

    default:
      /* Unknown file type */
      return DFCARD_WRONG_FILE_TYPE;
  }

  return DF_OPERATION_OK;
}

/* 
 * ChangeFileSettings
 * ------------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ChangeFileSettings(BYTE file_id, BYTE comm_mode, WORD access_rights)
{
  BOOL    comm_encrypted;
  SWORD   status;
  WORD    old_access_rights;

  /* Before being able to send the new file settings to the PICC, we need to know whether */
  /* the current access rights setting requires the communication to be encrypted.        */
  status = DesFireCard_GetFileSettings(file_id, NULL, NULL, &old_access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;
  comm_encrypted = (old_access_rights & DF_CHANGE_ACCESS_RIGHTS);

  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CHANGE_FILE_SETTINGS;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = comm_mode;
  DFC_buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  DFC_buflen = 5;

  if (comm_encrypted != 0x0E)
  {
    /* If encrypted communication is requested, append a CRC, computed over the communication mode         */
    /* and access rights arguments (3bytes), to the command string. Then 'Encrypt' these arguments and the */
    /* CRC to ensure that only an authenticated PCD can change the settings.                               */
    /* Note: since mifare DESFire always runs in DES ENcryption mode this is actually a DES                */
    /* DEcryption operation.                                                                               */
    SPROX_ComputeIso14443ACrc(&DFC_buffer[INF + 5], &DFC_buffer[INF + 2], 3);
    memset(&DFC_buffer[INF + 7], 0x00, 3);
    DesFire_xDES_Send(&DFC_buffer[INF + 2], 1, DFC_session_key);
    DFC_buflen += 5;
  }

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateStdDataFile
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateStdDataFile(BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD file_size)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CREATE_STD_DATA_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = comm_mode;
  DFC_buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  DFC_buffer[INF + 5] = (BYTE) (file_size & 0x00FF);
  file_size >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (file_size & 0x00FF);
  file_size >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (file_size & 0x00FF);
  DFC_buflen = 8;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateBackupDataFile
 * --------------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateBackupDataFile(BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD file_size)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CREATE_BACKUP_DATA_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = comm_mode;
  DFC_buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  DFC_buffer[INF + 5] = (BYTE) (file_size & 0x000000FF);
  file_size >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (file_size & 0x000000FF);
  file_size >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (file_size & 0x000000FF);
  DFC_buflen = 8;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateValueFile
 * ---------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateValueFile(BYTE file_id, BYTE comm_mode, WORD access_rights, SDWORD lower_limit, SDWORD upper_limit,
                                                                      SDWORD initial_value, BYTE limited_credit_enabled)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CREATE_VALUE_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = comm_mode;
  DFC_buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  DFC_buffer[INF + 5] = (BYTE) (lower_limit & 0x000000FF);
  lower_limit >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (lower_limit & 0x000000FF);
  lower_limit >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (lower_limit & 0x000000FF);
  lower_limit >>= 8;
  DFC_buffer[INF + 8] = (BYTE) (lower_limit & 0x000000FF);
  DFC_buffer[INF + 9] = (BYTE) (upper_limit & 0x000000FF);
  upper_limit >>= 8;
  DFC_buffer[INF + 10] = (BYTE) (upper_limit & 0x000000FF);
  upper_limit >>= 8;
  DFC_buffer[INF + 11] = (BYTE) (upper_limit & 0x000000FF);
  upper_limit >>= 8;
  DFC_buffer[INF + 12] = (BYTE) (upper_limit & 0x000000FF);
  DFC_buffer[INF + 13] = (BYTE) (initial_value & 0x000000FF);
  initial_value >>= 8;
  DFC_buffer[INF + 14] = (BYTE) (initial_value & 0x000000FF);
  initial_value >>= 8;
  DFC_buffer[INF + 15] = (BYTE) (initial_value & 0x000000FF);
  initial_value >>= 8;
  DFC_buffer[INF + 16] = (BYTE) (initial_value & 0x000000FF);
  DFC_buffer[INF + 17] = limited_credit_enabled;
  DFC_buflen = 18;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateLinearRecordFile
 * ----------------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateLinearRecordFile(BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD record_size,
                                                                             DWORD max_records)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CREATE_LINEAR_RECORD_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = comm_mode;
  DFC_buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  DFC_buffer[INF + 5] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (record_size & 0x000000FF);
  DFC_buffer[INF + 8] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  DFC_buffer[INF + 9] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  DFC_buffer[INF + 10] = (BYTE) (max_records & 0x000000FF);
  DFC_buflen = 11;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * CreateCyclicRecordFile
 * ----------------------
 */

SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CreateCyclicRecordFile(BYTE file_id, BYTE comm_mode, WORD access_rights, DWORD record_size,
                                                                             DWORD max_records)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_CREATE_CYCLIC_RECORD_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = comm_mode;
  DFC_buffer[INF + 3] = (BYTE) (access_rights & 0x00FF);
  access_rights >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (access_rights & 0x00FF);
  DFC_buffer[INF + 5] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (record_size & 0x000000FF);
  record_size >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (record_size & 0x000000FF);
  DFC_buffer[INF + 8] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  DFC_buffer[INF + 9] = (BYTE) (max_records & 0x000000FF);
  max_records >>= 8;
  DFC_buffer[INF + 10] = (BYTE) (max_records & 0x000000FF);
  DFC_buflen = 11;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * DeleteFile
 * ----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_DeleteFile(BYTE file_id)
{
  /* Create the info block containing the command code and the given parameters. */
  DFC_buffer[INF + 0] = DF_DELETE_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buflen = 2;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * ReadData
 * ---------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadData(BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD max_count, BYTE * data,
                                                               DWORD * done_count)
{
  return DesFireCard_ReadFile(DF_READ_DATA, file_id, comm_mode, from_offset, max_count, 1, data, done_count);
}

/*
 * ReadData2
 * ---------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadData2(BYTE file_id, DWORD from_offset, DWORD max_count, BYTE * data, DWORD * done_count)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the AccessRights field (settings r and r/w) we have to decide whether     */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If DFC_auth_key does neither match r nor r/w and one of this settings                  */
  /* contains the value "ever" (0x0E) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != DFC_auth_key) && (read_write_access != DFC_auth_key) && ((read_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_ReadData(file_id, comm_mode, from_offset, max_count, data, done_count);
}

/*
 * WriteData
 * ---------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteData(BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, BYTE * data)
{
  return DesFireCard_WriteFile(DF_WRITE_DATA, file_id, comm_mode, from_offset, size, data);
}

/*
 * WriteData2
 * ----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteData2(BYTE file_id, DWORD from_offset, DWORD size, BYTE * data)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the access_rights field (settings w and r/w) we have to decide whether   */
  /* we are able to communicate in the mode indicated by comm_mode.                        */
  /* If DFC_auth_key does neither match w nor r/w and one of this settings                 */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                   */
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((write_only_access != DFC_auth_key) && (read_write_access != DFC_auth_key) && ((write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_WriteData(file_id, comm_mode, from_offset, size, data);
}

/*
 * GetValue
 * --------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetValue(BYTE file_id, BYTE comm_mode, SDWORD * value)
{
  SDWORD  temp;
  SWORD   status;

  /* Create the info block containing the command code and the file ID argument. */
  DFC_buffer[INF + 0] = DF_GET_VALUE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buflen = 2;

  if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* The communication mode is DF_COMM_MODE_ENCIPHERED.                   */
    /* GetValue returns 9 bytes, the status byte and a cryptogram block     */
    /* containing the four byte value, 2 CRC bytes and two 0 padding bytes. */
    status = DesFireCard_Command(FALSE, 9, ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    /* Decipher the received block. */
    DesFire_xDES_Recv(&DFC_buffer[INF + 1], 1, DFC_session_key);

    if (SPROX_ComputeIso14443ACrc(NULL, &DFC_buffer[INF + 1], 4 + 2))
    {
      /* abortion due to integrity error -> wrong CRC */
      return DFCARD_WRONG_CRC;
    }

    /* Check also the padding bytes for enhanced security. */
    if (DFC_buffer[INF + 7] || DFC_buffer[INF + 8])
    {
      /* Error: cryptogram contains incorrect padding bytes. */
      return DFCARD_WRONG_PADDING;
    }
  } else if (comm_mode == DF_COMM_MODE_MACED)
  {
    BYTE    mac32[4];

    /* The communication mode is DF_COMM_MODE_MACED.                      */
    /* GetValue returns 9 bytes, the status byte, the four byte value and */
    /* the four byte MAC.                                                 */
    status = DesFireCard_Command(FALSE, 9, ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    /* Check the received MAC. */
    DesFire_xDES_MAC_b(&DFC_buffer[INF + 1], 4, DFC_session_key, mac32);
    if (memcmp(mac32, &DFC_buffer[INF + 5], 4))
    {
      /* abortion due to integrity error -> wrong MAC */
      return DFCARD_WRONG_MAC;
    }
  } else
  {
    /* GetValue returns 5 bytes, the status byte and the four byte value. */
    status = DesFireCard_Command(FALSE, 5, ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;
  }

  /* Return the requested value bytes. */
  temp = DFC_buffer[INF + 4];
  temp <<= 8;
  temp += DFC_buffer[INF + 3];
  temp <<= 8;
  temp += DFC_buffer[INF + 2];
  temp <<= 8;
  temp += DFC_buffer[INF + 1];
  if (value != NULL)
    *value = temp;

  /* Success. */
  return DF_OPERATION_OK;
}

/*
 * GetValue2
 * ---------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_GetValue2(BYTE file_id, SDWORD * value)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* GetValue is controlled by the fields r, w, r/w within the access_rights.               */
  /* Depending on the access_rights field (settings r, w and r/w) we have to decide whether */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If DFC_auth_key does neither match r, w nor r/w and one of this settings               */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless  */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != DFC_auth_key)
      && (write_only_access != DFC_auth_key)
      && (read_write_access != DFC_auth_key) && ((read_only_access == 0x0E) || (write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_GetValue(file_id, comm_mode, value);
}

/*
 * ReadRecords
 * -----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadRecords(BYTE file_id, BYTE comm_mode, DWORD from_record, DWORD max_record_count, DWORD record_size,
                                                                  BYTE * data, DWORD * record_count)
{
  BYTE    file_type;
  DWORD   done_size;
  DF_ADDITIONAL_FILE_SETTINGS file_settings;
  SWORD   status;

  memset(&file_settings, 0x00, sizeof(file_settings));

  /* if a pointer was passed for retrieving the number of records read    */
  /* and no record size was specified, we have to get the record size via */
  /* the GetFileSettings command                                          */
  if ((record_count != NULL) && (record_size == 0))
  {
    status = DesFireCard_GetFileSettings(file_id, &file_type, 0, 0, &file_settings);
    if (status != DF_OPERATION_OK)
      return status;

    /* if the file type indicates that this is not a record file */
    /* we are not able to proceed                                */
    if ((file_type != DF_LINEAR_RECORD_FILE) && (file_type != DF_CYCLIC_RECORD_FILE))
      return DFCARD_WRONG_FILE_TYPE;

    record_size = file_settings.stRecordFileSettings.eRecordSize;

    /* if eRecordSize becomes zero under any circumstances this     */
    /* would lead to a division by zero at the end of this function */
    /* thus making it impossible to proceed                         */
    if (record_size == 0)
      return DFCARD_WRONG_RECORD_SIZE;
  }

  /* Call the ReadFile function */
  status = DesFireCard_ReadFile(DF_READ_RECORDS, file_id, comm_mode, from_record, max_record_count, record_size, data, &done_size);
  if (status != DF_OPERATION_OK)
    return status;

  /* we have to check whether we received multiples of record_size   */
  /* if this is not the case this means that a format error occurred */
  if ((done_size % record_size) != 0)
    return DFCARD_WRONG_LENGTH;

  /* calculate the number of records if a pointer has been passed */
  if (record_count != NULL)
    *record_count = done_size / record_size;

  return DF_OPERATION_OK;
}

/*
 * ReadRecords2
 * ------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ReadRecords2(BYTE file_id, DWORD from_record, DWORD max_record_count, DWORD record_size, BYTE * data,
                                                                   DWORD * record_count)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the AccessRights field (settings r and r/w) we have to decide whether     */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If DFC_auth_key does neither match r nor r/w and one of this settings                  */
  /* contains the value "ever" (0x0E) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != DFC_auth_key) && (read_write_access != DFC_auth_key) && ((read_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_ReadRecords(file_id, comm_mode, from_record, max_record_count, record_size, data, record_count);
}

/*
 * WriteRecord
 * -----------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteRecord(BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, BYTE * data)
{
  return DesFireCard_WriteFile(DF_WRITE_RECORD, file_id, comm_mode, from_offset, size, data);
}

/*
 * WriteRecord2
 * ------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_WriteRecord2(BYTE file_id, DWORD from_offset, DWORD size, BYTE * data)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Depending on the access_rights field (settings w and r/w) we have to decide whether   */
  /* we are able to communicate in the mode indicated by comm_mode.                        */
  /* If DFC_auth_key does neither match w nor r/w and one of this settings                 */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                   */
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((write_only_access != DFC_auth_key) && (read_write_access != DFC_auth_key) && ((write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_WriteRecord(file_id, comm_mode, from_offset, size, data);
}

/*
 * LimitedCredit
 * -------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_LimitedCredit(BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  return DesFireCard_ModifyValue(DF_LIMITED_CREDIT, file_id, comm_mode, amount);
}

/*
 * LimitedCredit2
 * --------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_LimitedCredit2(BYTE file_id, SDWORD amount)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* LimitedCredit is controlled by the fields w and r/w within the access_rights.         */
  /* Depending on the access_rights field (settings w and r/w) we have to decide whether   */
  /* we are able to communicate in the mode indicated by comm_mode.                        */
  /* If DFC_auth_key does neither match w nor r/w and one of this settings                 */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless */
  /* of the mode indicated by comm_mode.                                                   */
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((write_only_access != DFC_auth_key) && (read_write_access != DFC_auth_key) && ((write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_LimitedCredit(file_id, comm_mode, amount);
}

/*
 * Credit
 * ------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Credit(BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  return DesFireCard_ModifyValue(DF_CREDIT, file_id, comm_mode, amount);
}

/*
 * Credit2
 * -------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Credit2(BYTE file_id, SDWORD amount)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Credit is controlled by the field r/w within the access_rights.                    */
  /* Depending on the AccessRights field (setting r/w) we have to decide whether        */
  /* we are able to communicate in the mode indicated by comm_mode.                     */
  /* If DFC_auth_key does not match r/w and this setting contains the value             */
  /* "ever" (0xE), communication has to be done in plain mode regardless of the mode    */
  /* indicated by comm_mode.                                                            */
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_write_access != DFC_auth_key) && (read_write_access == 0x0E))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_Credit(file_id, comm_mode, amount);
}

/*
 * Debit
 * -----
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Debit(BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  return DesFireCard_ModifyValue(DF_DEBIT, file_id, comm_mode, amount);
}

/*
 * Debit2
 * ------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_Debit2(BYTE file_id, SDWORD amount)
{
  BYTE    comm_mode;
  WORD    access_rights;
  BYTE    read_only_access;
  BYTE    write_only_access;
  BYTE    read_write_access;
  SWORD   status;

  /* we have to receive the communications mode first */
  status = DesFireCard_GetFileSettings(file_id, NULL, &comm_mode, &access_rights, NULL);
  if (status != DF_OPERATION_OK)
    return status;

  /* Debit is controlled by the fields r, w, r/w within the access_rights.                  */
  /* Depending on the access_rights field (settings r, w and r/w) we have to decide whether */
  /* we are able to communicate in the mode indicated by comm_mode.                         */
  /* If DFC_auth_key does neither match r, w nor r/w and one of this settings               */
  /* contains the value "ever" (0xE) communication has to be done in plain mode regardless  */
  /* of the mode indicated by comm_mode.                                                    */
  read_only_access = (BYTE) ((access_rights & DF_READ_ONLY) >> DF_GET_READ_ONLY);
  write_only_access = (BYTE) ((access_rights & DF_WRITE_ONLY) >> DF_GET_WRITE_ONLY);
  read_write_access = (BYTE) ((access_rights & DF_READ_WRITE) >> DF_GET_READ_WRITE);

  if ((read_only_access != DFC_auth_key)
      && (write_only_access != DFC_auth_key)
      && (read_write_access != DFC_auth_key) && ((read_only_access == 0x0E) || (write_only_access == 0x0E) || (read_write_access == 0x0E)))
  {
    comm_mode = DF_COMM_MODE_PLAIN;
  }

  /* Now execute the command */
  return DesFireCard_Debit(file_id, comm_mode, amount);
}

/*
 * ClearRecordFile
 * ---------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_ClearRecordFile(BYTE file_id)
{
  DFC_buffer[INF + 0] = DF_CLEAR_RECORD_FILE;
  DFC_buffer[INF + 1] = file_id;
  DFC_buflen = 2;

  /* response must be of one byte length if successfully terminated */
  return DesFireCard_Command(FALSE, 1, ACCEPT_OPERATION_OK);
}

/*
 * CommitTransaction
 * -----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_CommitTransaction(void)
{
  /* Create the info block containing the command code. */
  DFC_buffer[INF + 0] = DF_COMMIT_TRANSACTION;
  DFC_buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 * AbortTransaction
 * ----------------
 */
SPROX_DESFIRE_LIB SWORD SPROX_DESFIRE_API DesFireCard_AbortTransaction(void)
{
  /* Create the info block containing the command code. */
  DFC_buffer[INF + 0] = DF_ABORT_TRANSACTION;
  DFC_buflen = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 **************************************************************************************************
 *
 *  ReadFile / WriteFile / ModifyValue internal functions
 *
 *************************************************************************************************
 */

/*
 * ReadFile
 * --------
 */
static SWORD DesFireCard_ReadFile(BYTE read_command, BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD item_count, DWORD item_size, BYTE * data,
                                  DWORD * done_size)
{
  SWORD   status;
  DWORD   recv_len = 0;
  DWORD   padd_pos = 0;
  DWORD   byte_count;
  DWORD   temp;
#ifndef UNDER_CE
  static BYTE buffer[4096 + 16];
#else
  BYTE    buffer[4096 + 16];
#endif

  /* We have to calculate the number of bytes as this function works at byte granularity.
     Cast the DWORD to double in the multiplication to detect an overflow in the result.
     If an overflow occurs, continue with the maximum number provided by a DWORD. The PICC
     will refuse the read request in this case with a BOUNDARY_ERROR. */
  if (item_count * item_size > 4096)
    byte_count = 0xFFFFFFFF;
  else
    byte_count = item_count * item_size;

  DFC_buffer[INF + 0] = read_command;
  /* we do not have to take the item_size into account for the 
     generation of the command header as the PICC resolves the meaning of this
     parameter due to the specific command code */
  DFC_buffer[INF + 1] = file_id;
  temp = from_offset;
  DFC_buffer[INF + 2] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 3] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (temp & 0x000000FF);
  temp = item_count;
  DFC_buffer[INF + 5] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (temp & 0x000000FF);
  DFC_buflen = 8;

  for (;;)
  {
    status = DesFireCard_Command(TRUE, 0, DF_ADDITIONAL_FRAME | ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
    {
      return status;
    }
    memcpy(&buffer[recv_len], &DFC_buffer[INF + 1], DFC_buflen - 1);
    recv_len += (DFC_buflen - 1);
    if (DFC_buffer[INF + 0] != DF_ADDITIONAL_FRAME)
      break;
    DFC_buflen = 1;
  }

  /* correct termination of this loop only is possible if we received
     a frame with the status code sgbOPERATION_OK
     this means that this was the last frame sent by the PICC
     the additional frame status code which was prepared at the end of the loop 
     is not sent but discarded */

  /* decide upon the communications mode which cryptographic operation
     is to be applied on the received data */
  if ((comm_mode == DF_COMM_MODE_PLAIN) || (comm_mode == DF_COMM_MODE_PLAIN2))
  {
    if (item_count == 0)
    {
      /* if no length parameter has been passed, 
         eNBytes is set with the number of received bytes
         this is valid for both, record files and data files */
      byte_count = recv_len;
    } else
    {
      /* if a length parameter has been passed,
         we check, whether we got exactly that much data. */
      if (recv_len != byte_count)
      {
        /* length error */
        return DFCARD_WRONG_LENGTH;
      }
    }
  } else if (comm_mode == DF_COMM_MODE_MACED)
  {
    BYTE    mac32[4];

    DesFire_xDES_MAC_b(buffer, (WORD) (recv_len - 4), DFC_session_key, mac32);
    if (memcmp(mac32, &buffer[recv_len - 4], 4))
    {
      /* abortion due to integrity error -> wrong MAC */
      return DFCARD_WRONG_MAC;
    }
    if (item_count == 0)
    {
      /* if no length parameter has been passed 
         eNBytes is set with the number of received bytes
         this is valid for both, record files and data files */
      byte_count = recv_len - 4;
    }
  } else if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* check whether the received data count is multiple of 8 */
    if (recv_len % 8)
    {
      return DF_LENGTH_ERROR;
    }

    /* at first we have to decipher the buffer */
    DesFire_xDES_Recv(buffer, (WORD) (recv_len / 8), DFC_session_key);

    /* now there are two different paths caused by the two different padding methods */

    /* if the data length was specified in the parameter section of the command
       the buffer content is padded with all zeroes or possibly none,
       after the CRC to reach multiples of 8 bytes, suitable for the DesOperation 
       in the other case, where no data length was specified, the buffer content is padded
       at least with one byte coded to 0x80 which is followed by all zeroes or possibly 
       none, after the CRC to reach multiples of 8 bytes, suitable for the DesOperation */
    if (item_count != 0)
    {
      /* set the starting position for checking the zero padding bytes (after CRC) */
      padd_pos = byte_count + 2;
    } else
    {
      /* no length was specified, therefore the second padding rule applies
         we have to find the character 0x80 at the end of the received data
         and obtain the number of data bytes by subtracting 2 bytes from the position
         where 0x80 was found ( position + 1 = length ) */
      for (byte_count = (recv_len - 1); byte_count > 0; byte_count--)
      {
        if (buffer[byte_count] == 0x80)
        {
          byte_count = byte_count - 2;
          break;
        } else if (buffer[byte_count] != 0x00 || byte_count == 1)
        {
          return DFCARD_WRONG_PADDING;
        }
      }
      padd_pos = byte_count + 3;
    }

    /* this is the CRC check for both cases (specified and unspecified length) */
    if (SPROX_ComputeIso14443ACrc(NULL, buffer, (WORD) (byte_count + 2)))
    {
      /* abortion due to integrity error -> wrong CRC */
      return DFCARD_WRONG_CRC;
    }

    /*  Finally we do a check on the rest of the padding bytes ( if there are any )
       which have to be all zero for both cases */

    /* note: in case of the unspecified length parameter, there must be at least
       the 0x80 byte ( has been checked already! ) from there on
       both cases are again the same */
    if (recv_len > byte_count + 2)
    {
      for (; padd_pos < recv_len; padd_pos++)
      {
        if (buffer[padd_pos] != 0x00)
        {
          return DFCARD_WRONG_PADDING;
        }
      }
    }
  }

  /* return the number of bytes if a pointer was passed */
  if (done_size != NULL)
    *done_size = byte_count;

  /* copy data */
  if (data != NULL)
    memcpy(data, buffer, byte_count);

  return DF_OPERATION_OK;
}

/*
 * WriteFile
 * ---------
 */
static SWORD DesFireCard_WriteFile(BYTE write_command, BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, BYTE * data)
{
  DWORD   done_size;
  WORD    next_size;
  DWORD   temp;
  SWORD   status;
#ifndef UNDER_CE
  static BYTE buffer[4096 + 16];
#else
  BYTE    buffer[4096 + 16];
#endif

  memset(buffer, 0xCC, sizeof(buffer)); // TODO
  memcpy(buffer, data, size);

  done_size = 0;

  DFC_buffer[INF + 0] = write_command;
  DFC_buffer[INF + 1] = file_id;
  temp = from_offset;
  DFC_buffer[INF + 2] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 3] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (temp & 0x000000FF);
  temp = size;
  DFC_buffer[INF + 5] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 6] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  DFC_buffer[INF + 7] = (BYTE) (temp & 0x000000FF);
  DFC_buflen = 8;

  /* decide upon the communications mode which cryptographic */
  /* operation is to be applied on the data                  */

  if (comm_mode == DF_COMM_MODE_MACED)
  {
    /* append the 4 bytes MAC */
    DesFire_xDES_MAC_b(buffer, (WORD) size, DFC_session_key, &buffer[size]);
    size += 4;
  } else if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* at first we have to append the 2 bytes CRC */
    SPROX_ComputeIso14443ACrc(&buffer[size], buffer, (WORD) size);
    size += 2;

    while (size % 8)
      buffer[size++] = 0;

    /* finally we do the cipher operation on the modified data */
    DesFire_xDES_Send(buffer, (WORD) (size / 8), DFC_session_key);
  }

  /* if comm_mode is neither MACed nor ciphered we leave the data as it is */
  /* this means a plain communication                                      */

  /* Now our data is ready for being sent to the PICC */

  do
  {
    /* determine the limiting factor for the number of bytes to be appended to     */
    /* the DFC_buffer                                                              */
    /* this is either the maximum frame size or the number of bytes which are left */
    if ((size - done_size) > (DWORD) (DF_MAX_INFO_FRAME_SIZE - DFC_buflen))
    {
      /* transfer the calculated number of bytes  */
      next_size = (WORD) (DF_MAX_INFO_FRAME_SIZE - DFC_buflen);
    } else
    {
      /* transfer all the bytes left  */
      next_size = (WORD) (size - done_size);
    }

    /* Now put these bytes into the DFC_buffer */
    memcpy(&DFC_buffer[INF + DFC_buflen], &buffer[done_size], next_size);
    DFC_buflen += (WORD) next_size;

    /* only one byte response is allowed                                */
    /* allowed status codes are DF_ADDITIONAL_FRAME and DF_OPERATION_OK */
    status = DesFireCard_Command(FALSE, 1, DF_ADDITIONAL_FRAME | ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
    {
      return status;
    }

    done_size += (DWORD) next_size;

    /* put the ADDITONAL_FRAME status code for the next frame into the buffer */
    DFC_buffer[INF + 0] = DF_ADDITIONAL_FRAME;
    DFC_buflen = 1;
  } while (done_size < size);

  /* because ( eNumOfBytesExtracted = eNBytesToWrite ) is the only way for     */
  /* leaving the loop correctly an interrupted write operation is detected via */
  /* a status code different from DF_OPERATION_OK                              */
  return status;
}

/*
 * ModifyValue
 * -----------
 */
static SWORD DesFireCard_ModifyValue(BYTE modify_command, BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  /* Create the info block containing the command code, the file ID argument and the */
  /* modification amount argument.                                                   */
  DFC_buffer[INF + 0] = modify_command;
  DFC_buffer[INF + 1] = file_id;
  DFC_buffer[INF + 2] = (BYTE) (amount & 0x000000FF);
  amount >>= 8;
  DFC_buffer[INF + 3] = (BYTE) (amount & 0x000000FF);
  amount >>= 8;
  DFC_buffer[INF + 4] = (BYTE) (amount & 0x000000FF);
  amount >>= 8;
  DFC_buffer[INF + 5] = (BYTE) (amount & 0x000000FF);
  DFC_buflen = 6;

  /* Process the amount argument according to the selected communication mode. */
  if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* Append a CRC to the amount argument. */
    SPROX_ComputeIso14443ACrc(&DFC_buffer[INF + 6], &DFC_buffer[INF + 2], 4);

    /* 'Encrypt' the amount argument bytes and the CRC bytes padding         */
    /* the DES block with '0' bytes.                                         */
    /* Note: since mifare DESFire always runs in DES ENcryption mode this is */
    /* actually a DES DEcryption operation.                                  */
    DFC_buffer[INF + 8] = 0;
    DFC_buffer[INF + 9] = 0;

    DesFire_xDES_Send(&DFC_buffer[INF + 2], 1, DFC_session_key);
    DFC_buflen += 4;
  } else if (comm_mode == DF_COMM_MODE_MACED)
  {
    /* Append a MAC to the amount argument */
    DesFire_xDES_MAC_b(&DFC_buffer[INF + 2], 4, DFC_session_key, &DFC_buffer[INF + 6]);
    DFC_buflen += 4;
  }

  /* Communicate the info block to the card and check the operation's return status. */
  return DesFireCard_Command(FALSE, 0, ACCEPT_OPERATION_OK);
}

/*
 **************************************************************************************************
 *
 *  Command : "low" level command<->response transfert
 *
 *************************************************************************************************
 */

static SWORD DesFireCard_Command(BOOL fast, WORD accept_len, BYTE accept_status)
{
  SWORD   status;
  BYTE    card_status;
  WORD    len = sizeof(DFC_buffer);

//  D(printf("S: "));
//  D(dump_buffer(DFC_buffer, DFC_buflen);
//  D(printf("\n"));

  /* Communicate the prepared buffer to the PICC and get its response. */
  if (fast)
  {
    status = SPROX_TclA_ExchangeDF(DFC_cid, DFC_buffer, DFC_buflen, DFC_buffer, &len);
  } else
  {
    status = SPROX_TclA_Exchange(5, DFC_cid, 0xFF, DFC_buffer, DFC_buflen, DFC_buffer, &len);
  }
  if (status != MI_OK)
  {
    /* Error: could not exchange info blocks with the PICC. */
    return status;
  }

  /* Remember received size */
  DFC_buflen = len;

//  D(printf("R: "));
//  D(dump_buffer(DFC_buffer, DFC_buflen);
//  D(printf("\n"));

  /* Check whether the PICC's response consists of at least one byte.
     If the response is empty, we can not even determine the PICC's status.
     Therefore an empty response is always a length error. */
  if (len < 1)
  {
    /* Error: block with inappropriate number of bytes received from the PICC. */
    return DFCARD_WRONG_LENGTH;
  }

  /*  Check the status byte received from the PICC.
     Depending on which of the flags (sgbACCEPT_OPERATION_OK and sgbACCEPT_ADDITIONAL_FRAME)
     are set in the bExpectedStatusFlags parameter, we accept either the status byte 
     sgbOPERATION_OK or the status byte sgbDF_ADDITIONAL_FRAME or both of them. */
  card_status = DFC_buffer[INF + 0];

  if (!(((card_status == DF_OPERATION_OK) && (accept_status & ACCEPT_OPERATION_OK)) ||
        ((card_status == DF_ADDITIONAL_FRAME) && (accept_status & ACCEPT_ADDITIONAL_FRAME))))
  {
    /* Error: error or unexpected status received from the PICC. */
    return (DFCARD_ERROR - card_status);
  }

  /* Finally check the length of the response. It must be > 0 (already checked) and
     <= sgbMAX_INFO_FRAME_SIZE in all cases.
     If bExpectedResponseLength is != 0, we expect a response with exactly that length.
     If bExpectedResponseLength == 0, we allow any response length > 0 and 
     <= sgbMAX_INFO_FRAME_SIZE. */
  if (((!fast) && (len > MAX_INFO_FRAME_SIZE)) || ((accept_len != 0) && (accept_len != len)))
  {
    /* Error: block with inappropriate number of bytes received from the PICC. */
    return DFCARD_WRONG_LENGTH;
  }

  /* Return the status code from the PICC without any prefix code. */
  return status;
}
