#include "springprox.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef WIN32
  #include <conio.h> /* For getch */
#endif

#include "fingertk.h"


/* Display the application menu and process user's entries */
/* ------------------------------------------------------- */

#define EMPTY_DATA      0 /* No data available yet */
#define RETRIEVED_DATA  1 /* Data have been read from a tag, file, or other media, they can be verified but not save */
#define ENROLLED_DATA   2 /* Data have been collected through the user enrollment process, they can be saved to various media */

void pause(void);

/* Displays the menu and waits for a valid entry */
char menu(int data_status, FINGERTK_DATA_ST *data)
{
  char   buffer[40+1];
  char   allowed[40] = " ";
  char   key;
  size_t i;

  printf("\n");
  printf("Pro-Active FingerTK\n");
  printf("-------------------\n");
  printf("\n");

  if (data_status >= RETRIEVED_DATA)
  {
    printf("Active user:\n");

    strncpy(buffer, data->id, 20);
    buffer[20] = '\0';
    printf("\tID         = %s\n", buffer);

    strncpy(buffer, data->lastname, 40);
    buffer[40] = '\0';
    printf("\tLast name  = %s\n", buffer);

    strncpy(buffer, data->firstname, 40);
    buffer[40] = '\0';
    printf("\tFirst name = %s\n", buffer);

    if ((data->finger1_len == 0) && (data->finger2_len == 0))
    {
      printf("\t!!! No finger recorded !!!\n");
    } else
    {
      if ((data->finger1_len > 0) && (data->finger2_len > 0))
      {
        printf("\t2 fingers recorded\n");
      } else
      {
        printf("\t1 finger recorded\n");
      }
      printf("\n");
      printf("(V)\tVerify user's finger(s)\n");
      strcat(allowed, "v");
      printf("\n");
      if (data_status >= ENROLLED_DATA)
      {
        printf("(S)\tSave user's data to a file\n");
        printf("(W)\tWrite user's data to a new tag\n");
        printf("   \t       (Pro-Active format, Mifare or Desfire)\n");
        strcat(allowed, "sw");
        printf("\n");
      }
    }
  } else
  {
    printf("No active user\n");
  }

  printf("\n");
  printf("(E)\tEnroll a new user\n");
  printf("\n");

  printf("Load user's data :\n\n");
  printf("(L)\tLoad from file\n");
  printf("(R)\tRead a Pro-Active formatted tag (Mifare or Desfire)\n");
  strcat(allowed, "lr");
  printf("\n");
  strcat(allowed, "gt");

  printf("\n");

  printf("(X)\tExit this application\n");
  strcat(allowed, "ex");

  printf("Ca commence ici\n");

  for (;;)
  {
#ifdef WIN32
    while (!kbhit());
    key = getch();
#else
    scanf("%c", &key);
#endif
    key = tolower(key);
    for (i=0; i<strlen(allowed); i++)
      if (allowed[i] == key)
        return key;
  }
}

void interactive(void)
{
  FINGERTK_DATA_ST data;
  int              data_status = EMPTY_DATA;
  char             key;

  for(;;)
  {
    key = menu(data_status, &data);
    switch (key)
    {
      case 'v' : verify(&data);
                 break;

      case 'e' : if (enroll(&data))
                   data_status = ENROLLED_DATA;
                 else
                   data_status = EMPTY_DATA;
                 break;

      case 's' : save_to_file(&data);
                 break;

      case 'w' : write_to_tag(&data);
                 break;

      case 'l' : if (load_from_file(&data))
                   data_status = ENROLLED_DATA;
                 else
                   data_status = EMPTY_DATA;
                 break;

      case 'r' : if (read_from_tag(&data))
                   data_status = RETRIEVED_DATA;
                 else
                   data_status = EMPTY_DATA;
                 break;

      case 'x' : printf("Exiting...\n");
                 return;
    }

    if (key != ' ')
      pause();
  }
}

/* Entry point of the application */
/* ------------------------------ */

int main(int argc, char **argv)
{
  char  buffer[64];
  char  mso_product[256];
  char  mso_firmware[256];
  char  mso_sensor[256];
  SWORD rc;

  printf("Pro-Active SpringBIO Developer's Kit\n");
  printf("\n");
  printf("SpringBIO ANSI C sample projet 'fingertk'\n");
  printf("-----------------------------------------\n");
  printf("\n");
  printf("Copyright Pro-Active 2005\n");

  rc = SPROX_GetLibrary(buffer, 64);
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", buffer);
  printf("\n");


  /* Open the reader */
  /* --------------- */
  if (argc >= 2)
    rc = SPROX_ReaderOpen(argv[1]);
  else
    rc = SPROX_ReaderOpen(NULL);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  rc = SPROX_ReaderGetDevice(buffer, 64);
  if (rc == MI_OK)
    printf("Reader found on %s\n", buffer);

  rc = SPROX_ReaderGetFirmware(buffer, 64);
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", buffer);

  if (0)
  {
    WORD len = 16;
    for (;;)
    {
      rc = SPROX_Echo(len);
      if (rc == MI_OK)
        printf("%d -> OK\n", len);
      else
      {
        printf("%d -> KO %d\n", len, rc);
        break;
      }
      len += 16;
    }
  }

  /* Check that the reader has an MSO fingerprint reader on board */
  /* ------------------------------------------------------------ */

  rc = SPROX_MSO_Open(mso_product, mso_firmware, mso_sensor);
  if (rc == MI_NY_IMPLEMENTED)
  {
    printf("This device doesn't have a fingerprint reader !\n");
    rc = MI_OK;
    goto close;
  } else
  if (rc != MI_OK)
  {
    printf("Fingerprint reader error %d\n", rc);
    goto close;
  }

  printf("Fingerprint reader found:\n");
  printf("Product  =\n%s\n", mso_product);
  printf("Firmware =\n%s\n", mso_firmware);
  printf("Sensor   =\n%s\n", mso_sensor);

  /* Transfert control to the menu loop */
  /* ---------------------------------- */

  interactive();

close:
  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_MSO_Close();
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc == MI_OK)
  {
    printf("Done\n");
  } else
  {
    printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }

  return 0;
}

void pause(void)
{
#ifdef WIN32
  printf("\n-- Hit a key --\n");
  getch();
#else
  char dummy[40];
  printf("\n-- Hit enter --\n");
  fgets(dummy, sizeof(dummy), stdin);
#endif
}
