/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2005 Pro-Active
  See COPYING for copyright and legal info 
   
  Ref_Calypso.c
  -------------

  This is the reference applications that shows how to retrieve Calypso
  card's identifier.

  It performs also ISO 14443-A and ISO 14443-B polling, so that we can
  study the side-effects of switching between different configs.
  Clearly we see that going to ISO 14443-A resets the Calypso card.
  
  JDA 14/11/2005 : initial release

*/

#include "springprox.h"
#include "sprox_calypso.h"

#ifdef WIN32
#include <conio.h>              /* For kbhit and getch */
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif

/*
 * Calypso GetUID function :
 * - sends APGEN and check REPGEN
 * - if picc_stop is TRUE, we send also ATTRIB and DISC to halt
 *   the Calypso card.
 */
SWORD SPROX_Calypso_GetUID(BYTE * picc_uid, BOOL picc_stop);

#ifndef WIN32
void onexit(void)
{
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();
}
#endif

int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  BYTE buffer[64];
  BYTE uid[12];
  BYTE uidlen;
  BYTE mode = 0;

  char *device = NULL;

  if (argc > 1)
    device = argv[1];

  printf("Pro-Active K531/SpringProx/CSB Developer's Kit\n");
  printf("\n");
  printf("Calypso polling demo software\n");
  printf("-----------------------------\n");
  printf("\n");
  printf("Copyright Pro-Active 2005, 2006\n");

  rc = SPROX_GetLibrary(buffer, 64);
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", buffer);
  printf("\n");

  /* Open reader */
  /* ----------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }
#ifndef WIN32
  atexit(onexit);
#endif


  rc = SPROX_ReaderGetDevice(buffer, 64);
  if (rc == MI_OK)
    printf("Reader found on %s\n", buffer);

  rc = SPROX_ReaderGetFirmware(buffer, 64);
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", buffer);

  /* RF field ON */
  SPROX_ControlRF(TRUE);

#ifdef WIN32
  printf("Waiting for tags... (Press any key to exit)\n\n");
  while (!kbhit())
#else
  printf("Waiting for tags... (Press <Ctrl>+C to exit)\n\n");
  for (;;)
#endif
  {
    /* Not too often... */
#ifdef WIN32
    Sleep(50);
#endif
#ifdef __linux
    usleep(5000);
#endif

    if (mode == 0)
    {
      /* Calypso polling */
      /* --------------- */

      /* Configure reader Calypso cards operation */
      rc = SPROX_ConfigureForCalypso();
      if (rc != MI_OK)
      {
        printf("Failed to configure the reader for Calypso operation\n");
        goto close;
      }

      /* Perform Calypso "APGEN" function (and keep only the UID) */
      rc = SPROX_Calypso_GetUID(uid, FALSE);
      if (rc == MI_OK)
      {
        printf("Calypso card found: UID=");
        for (i = 0; i < 4; i++)
          printf("%02X", uid[i]);
        printf("\n");
      }
    } else if (mode == 1)
    {
      /* ISO 14443-A polling */
      /* ------------------- */
      rc = SPROX_SetConfig(CFG_MODE_ISO_14443_A);
      if (rc != MI_OK)
      {
        printf
          ("\nFailed to configure the reader for ISO 14443-A operation (%d)\n",
           rc);
        goto close;
      }

      /* request+anticollision+select loop */
      for (;;)
      {
        uidlen = 12;
        rc = Mf500PiccActivateIdle(0, NULL, NULL, uid, &uidlen);
        if (rc == MI_NOTAGERR)
          break;                /* No more tag (or no tag at all) */
        if (rc != MI_OK)
        {
          /* Other error */
          printf("\nISO 14443-A activation error (%d)\n", rc);
          goto close;
        }

        printf("Tag (A) found: UID=");
        for (i = 0; i < uidlen; i++)
          printf("%02X", uid[i]);
        printf("\n");

        /* Halt currently active tag */
        rc = Mf500PiccHalt();
        if (rc != MI_OK)
        {
          printf("\nFailed to halt the ISO 14443-A tag (%d)\n", rc);
          goto close;
        }
      }
    } else if (mode == 2)
    {
      /* ISO 14443-B polling */
      /* ------------------- */

      rc = SPROX_SetConfig(CFG_MODE_ISO_14443_B);
      if (rc != MI_OK)
      {
        printf
          ("\nFailed to configure the reader for ISO 14443-B operation (%d)\n",
           rc);
        goto close;
      }

      rc = SPROX_TclB_ActivateIdle(0x00, uid);
      if (rc == MI_NOTAGERR)
      {
        /* No more tag (or no tag at all) */
      } else if (rc != MI_OK)
      {
        /* Other error */
        printf("\nISO 14443-B activation error (%d)\n", rc);
        goto close;
      } else
      {
        printf("Tag (B) found: UID=");
        for (i = 0; i < 4; i++)
          printf("%02X", uid[i]);
        printf("\n");

        /* Halt currently active tag */
        rc = SPROX_TclB_Halt(NULL);
        if (rc != MI_OK)
        {
          printf("\nFailed to halt the ISO 14443-B tag (%d)\n", rc);
          goto close;
        }
      }
    }

    mode++;
    mode %= 3;
  }

close:
  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc == MI_OK)
  {
    printf("Done\n");
  } else
  {
    printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }
  return 0;
}


SWORD SPROX_Calypso_GetUID(BYTE * picc_uid, BOOL picc_stop)
{
  SWORD rc;

  rc = SPROX_Calypso_APGEN(0x00, TRUE, TRUE, 0xFF, NULL, NULL, picc_uid);
  if ((rc == 0) && (picc_stop))
  {
    rc = SPROX_Calypso_ATTRIB(10, 10, picc_uid);
    if (rc == 0)
    {
      rc = SPROX_Calypso_DISC(10, 10);
    }
  }

  return rc;
}
